# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import datetime
import os
import sys

from cdpcli.compat import json
from cdpcli.compat import six
from cdpcli.exceptions import ParamValidationError
from cdpcli.model import ServiceModel
from cdpcli.parser import ResponseParserFactory
from cdpcli.serialize import create_serializer
from cdpcli.serialize import Serializer
import dateutil.parser
from tests import unittest
import yaml

UTF8 = 'utf-8'
PROTOCOL_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'protocol')


class TestSerializer(unittest.TestCase):

    def setUp(self):
        self.model = yaml.safe_load(open(os.path.join(PROTOCOL_DIR, 'service.yaml')))
        self.service_model = ServiceModel(self.model, 'servicename')
        self.serializer = create_serializer()
        self.non_validatring_serializer = Serializer()

    def _validate_request(self, request, url_path, params):
        self.assertEqual(len(request), 4)
        self.assertTrue('headers' in request)
        headers = request['headers']
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers['Content-Type'], 'application/json')
        self.assertEqual(request['url_path'], url_path)
        self.assertEqual(request['body'], json.dumps(params).encode(UTF8))

    def test_no_params(self):
        params = {}
        operation_model = self.service_model.operation_model('createDirector')
        request = self.non_validatring_serializer.serialize_to_request(
            params, operation_model)
        self._validate_request(request, '/directors/createDirector', params)
        with self.assertRaises(ParamValidationError):
            self.serializer.serialize_to_request(params, operation_model)

    def test_param_with_none_value(self):
        params = {}
        params['name'] = None
        operation_model = self.service_model.operation_model('createDirector')
        request = self.non_validatring_serializer.serialize_to_request(
            params, operation_model)
        self._validate_request(request, '/directors/createDirector', {})
        with self.assertRaises(ParamValidationError):
            self.serializer.serialize_to_request(params, operation_model)

    def test_string_param(self):
        params = {}
        params['name'] = 'test-name'
        operation_model = self.service_model.operation_model('createDirector')
        request = self.serializer.serialize_to_request(params, operation_model)
        self._validate_request(request, '/directors/createDirector', params)

    def test_list_param(self):
        params = {}
        params['names'] = ['test-name-1', 'test-name-2']
        operation_model = self.service_model.operation_model('describeDirectors')
        request = self.serializer.serialize_to_request(params, operation_model)
        self._validate_request(request, '/directors/describeDirectors', params)

    def test_nested_objects(self):
        params = {}
        params['inner'] = {'name': 'test-name'}
        operation_model = self.service_model.operation_model('nestedObjects')
        request = self.serializer.serialize_to_request(params, operation_model)
        self._validate_request(request, '/directors/nestedObjects', params)

    def test_array_of_objects(self):
        params = {}
        params['directors'] = [{'name': 'test-name-1',
                                'crn': 'test-crn-1'},
                               {'name': 'test-name-2',
                                'crn': 'test-crn-1'}]
        operation_model = self.service_model.operation_model('arrayOfObjects')
        request = self.serializer.serialize_to_request(params, operation_model)
        self._validate_request(request, '/directors/arrayOfObjects', params)

    if sys.version_info.major >= 3:  # bytes is introduced in python3
        def test_blob_param_bytes(self):
            params = {}
            params['data'] = b'Hello World'
            operation_model = self.service_model.operation_model('blobObject')
            request = self.serializer.serialize_to_request(params, operation_model)
            params['data'] = 'SGVsbG8gV29ybGQ='
            self._validate_request(request, '/directors/blobObject', params)

    def test_blob_param_bytearray(self):
        params = {}
        params['data'] = bytearray(b'Hello World')
        operation_model = self.service_model.operation_model('blobObject')
        request = self.serializer.serialize_to_request(params, operation_model)
        params['data'] = 'SGVsbG8gV29ybGQ='
        self._validate_request(request, '/directors/blobObject', params)

    def test_blob_param_base64(self):
        params = {}
        params['data'] = 'SGVsbG8gV29ybGQ='
        operation_model = self.service_model.operation_model('blobObject')
        request = self.serializer.serialize_to_request(params, operation_model)
        self._validate_request(request, '/directors/blobObject', params)


class TestParser(unittest.TestCase):

    def setUp(self):
        self.model = yaml.safe_load(open(os.path.join(PROTOCOL_DIR, 'service.yaml')))
        self.service_model = ServiceModel(self.model, 'servicename')
        self.parser = ResponseParserFactory().create_parser()

    def _create_response(self, body):
        response = {}
        response["status_code"] = 200
        response["headers"] = {}
        response["body"] = json.dumps(body).encode(UTF8)
        return response

    def _create_error_response(self, code, message):
        body = {}
        body['code'] = code
        body['message'] = message
        response = {}
        response["status_code"] = 301
        response["body"] = json.dumps(body).encode(UTF8)
        return response

    def _create_error_response_2(self, messages):
        body = {}
        body['errorMessages'] = messages
        response = {}
        response["status_code"] = 401
        response["body"] = json.dumps(body).encode(UTF8)
        return response

    def _create_error_response_3(self, message):
        body = {}
        body['error'] = message
        response = {}
        response["status_code"] = 400
        response["body"] = json.dumps(body).encode(UTF8)
        return response

    def _create_error_response_4(self, message):
        response = {}
        response["status_code"] = 409
        response["body"] = message.encode(UTF8)
        return response

    def _assert_same(self, v1, v2, path=""):
        if type(v1) is dict:
            for k in v1.keys():
                self.assertTrue(k in v2)
                self._assert_same(v1[k], v2[k])
            for k in v2.keys():
                self.assertTrue(k in v1)
                self._assert_same(v1[k], v2[k])
            return

        if type(v1) is datetime.datetime and isinstance(v2, six.text_type):
            v2 = dateutil.parser.parse(v2)
        if type(v2) is datetime.datetime and isinstance(v1, six.text_type):
            v1 = dateutil.parser.parse(v1)
        self.assertEquals(v1, v2)

    def _validate_parsed_response(self, response, parsed_response):
        self._assert_same(json.loads(response['body'].decode(UTF8)),
                          parsed_response)

    def _validate_error_response(self, response, parsed_response):
        self._assert_same(json.loads(response['body'].decode(UTF8)),
                          parsed_response['error'])

    def test_no_shape(self):
        director = {}
        director['name'] = 'test-name'
        director['crn'] = 'test-crn'
        director['endpoint'] = {'address': '1.2.3.4', 'port': 12345}
        response = self._create_response({'director': director})
        parsed_response = self.parser.parse(response, None)
        self.assertEqual({}, parsed_response)

    def test_no_response(self):
        response = {}
        response["status_code"] = 200
        response["headers"] = {}
        response["body"] = ""
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self.assertEqual({}, parsed_response)
        response["body"] = None
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self.assertEqual({}, parsed_response)
        response["body"] = {}
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self.assertEqual({}, parsed_response)

    def test_non_json_response(self):
        response = {}
        response["status_code"] = 404
        response["headers"] = {}
        response["body"] = "Page not found".encode(UTF8)
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self.assertEqual(parsed_response,
                         {'error': {'code': 'UNKNOWN_ERROR',
                                    'message': 'Page not found'}})

    def test_response(self):
        director = {}
        director['name'] = 'test-name'
        director['crn'] = 'test-crn'
        director['endpoint'] = {'address': '1.2.3.4', 'port': 12345}
        director['creationDate'] = "2014-07-31T12:26:00.000-07:00"
        director['booleanData'] = True
        director['numberData'] = "123.456"
        response = self._create_response({'director': director})
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self._validate_parsed_response(response, parsed_response)

    def test_invalid_datetime(self):
        director = {}
        director['creationDate'] = "BAD!-2014-07-31T12:26:00.000-07:00-BAD!"
        response = self._create_response({'director': director})
        operation_model = self.service_model.operation_model('createDirector')
        with self.assertRaises(ValueError):
            self.parser.parse(response, operation_model.output_shape)

    def test_response_with_array(self):
        director1 = {}
        director1['name'] = 'test-name-1'
        director1['crn'] = 'test-crn-1'
        director1['endpoint'] = {'address': '1.2.3.4', 'port': 12345}
        director2 = {}
        director2['name'] = 'test-name-2'
        director2['crn'] = 'test-crn-2'
        director2['endpoint'] = {'address': '2.3.4.5', 'port': 23456}
        response = self._create_response({'directors': [director1, director2]})
        operation_model = self.service_model.operation_model('describeDirectors')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self._validate_parsed_response(response, parsed_response)

    def test_response_with_missing_member(self):
        director = {}
        director['crn'] = 'test-crn'
        director['endpoint'] = {'address': '1.2.3.4', 'port': 12345}
        response = self._create_response({'director': director})
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self._validate_parsed_response(response, parsed_response)

    def test_response_with_none_member(self):
        director = {}
        director['name'] = None
        director['crn'] = 'test-crn'
        director['endpoint'] = {'address': '1.2.3.4', 'port': 12345}
        response = self._create_response({'director': director})
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        del director['name']
        response = self._create_response({'director': director})
        self._validate_parsed_response(response, parsed_response)

    def test_error(self):
        response = self._create_error_response(
            "YouMessedUpException",
            "You, sir, have messed up.")
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self._validate_error_response(response, parsed_response)

    def test_error_2(self):
        response = self._create_error_response_2([
            "YouMessedUpException",
            "You, sir, have messed up."])
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self.assertEquals({
            'error': {
                'code': '',
                'message': 'YouMessedUpException You, sir, have messed up.'}},
            parsed_response)

    def test_error_3(self):
        response = self._create_error_response_3(
            "YouMessedUpException "
            "You, sir, have messed up.")
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self.assertEquals({
            'error': {
                'code': '',
                'message': 'YouMessedUpException You, sir, have messed up.'}},
            parsed_response)

    def test_error_4(self):
        response = self._create_error_response_4(
            "YouMessedUpException "
            "You, sir, have messed up.")
        operation_model = self.service_model.operation_model('createDirector')
        parsed_response = self.parser.parse(response, operation_model.output_shape)
        self.assertEquals({
            'error': {
                'code': 'UNKNOWN_ERROR',
                'message': 'YouMessedUpException You, sir, have messed up.'}},
            parsed_response)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections import defaultdict

from cdpcli import ADDITIONAL_PROPERTIES
from cdpcli import LIST_TYPE
from cdpcli import MAP_TYPE
from cdpcli import OBJECT_TYPE
from cdpcli import REF_KEY
from cdpcli import REF_NAME_PREFIX
from cdpcli import TYPE_KEY
from cdpcli.compat import OrderedDict
from cdpcli.utils import CachedProperty
from cdpcli.utils import instance_cache


class NoShapeFoundError(Exception):
    pass


class InvalidShapeError(Exception):
    pass


class InvalidModelError(Exception):
    pass


class OperationNotFoundError(Exception):
    pass


def _get_shape_type(shape_data):
    shape_type = shape_data['type']
    if shape_type == 'string':
        if shape_data.get('format', None) == 'date-time':
            return 'datetime'
        if shape_data.get('format', None) == 'byte':
            return 'blob'
    if shape_type == 'number':
        return shape_data.get('format')
    if shape_type == 'object':
        if ADDITIONAL_PROPERTIES in shape_data:
            return 'map'
    return shape_type


class Shape(object):

    def __init__(self, name, shape_data, shape_resolver):
        self.name = name
        self.type_name = _get_shape_type(shape_data)
        # Array content definitions do not have description entries since the
        # array itself is documented.
        self.documentation = shape_data.get('description', "")
        self._shape_data = shape_data
        self._shape_resolver = shape_resolver
        self._cache = {}

    @CachedProperty
    def required_members(self):
        return self._shape_data.get('required', [])

    @CachedProperty
    def min_length(self):
        return self._shape_data.get('minLength', None)

    @CachedProperty
    def max_length(self):
        return self._shape_data.get('maxLength', None)

    @CachedProperty
    def minimum(self):
        return self._shape_data.get('minimum', None)

    @CachedProperty
    def maximum(self):
        return self._shape_data.get('maximum', None)

    @CachedProperty
    def is_paging_input_token(self):
        return self._shape_data.get('x-paging-input-token', False)

    @CachedProperty
    def is_paging_output_token(self):
        return self._shape_data.get('x-paging-output-token', False)

    @CachedProperty
    def is_paging_result(self):
        return self._shape_data.get('x-paging-result', False)

    @CachedProperty
    def is_page_size(self):
        return self._shape_data.get('x-paging-page-size', False)

    @CachedProperty
    def is_no_paramfile(self):
        return self._shape_data.get('x-no-paramfile', False)

    @CachedProperty
    def is_undocumented(self):
        return self._shape_data.get('x-deprecated', False)

    @CachedProperty
    def form_factors(self):
        if "x-form-factors" in self._shape_data:
            return self._shape_data['x-form-factors'].split(',')
        return None

    def _get_shape(self, name, shape_data):
        if REF_KEY in shape_data:
            # Reference value type
            # $ref: '#/definitions/SomeObject'
            return self._shape_resolver.resolve_shape_ref(name,
                                                          shape_data[REF_KEY])
        elif TYPE_KEY in shape_data:
            # Explicit value type, e.g., SomeShapeType
            return self._shape_resolver.get_shape(name, shape_data)
        else:
            raise InvalidShapeError("Unknown %s content: %s"
                                    % shape_data, self.name)


class ObjectShape(Shape):
    @CachedProperty
    def members(self):
        members = OrderedDict()
        for name, shape_data in self._shape_data.get('properties', {}).items():
            members[name] = self._get_shape(name, shape_data)
        return members


class MapShape(Shape):
    @CachedProperty
    def key(self):
        return self._get_shape('key', {'type': 'string'})

    @CachedProperty
    def value(self):
        additional_properties = self._shape_data.get(ADDITIONAL_PROPERTIES)
        # Map with Freeform object
        if type(additional_properties) == bool or not additional_properties:
            raise InvalidShapeError("Freeform maps are not supported: %s"
                                    % self.name)
        else:
            return self._get_shape('value', additional_properties)


class ArrayShape(Shape):
    @CachedProperty
    def member(self):
        return self._get_shape("items", self._shape_data['items'])


class StringShape(Shape):
    @CachedProperty
    def enum(self):
        return self._shape_data.get('enum', [])

    @CachedProperty
    def supported_options(self):
        return self._shape_data.get('x-supported-options', [])


class ShapeResolver(object):

    def __init__(self, definitions):
        self._definitions = definitions

    def get_shape(self, name, shape_data):
        try:
            shape_type = _get_shape_type(shape_data)
            if shape_type == OBJECT_TYPE:
                shape_cls = ObjectShape
            elif shape_type == 'map':
                shape_cls = MapShape
            elif shape_type == LIST_TYPE:
                shape_cls = ArrayShape
            elif shape_type == 'string':
                shape_cls = StringShape
            elif shape_type in ['integer',
                                'float',
                                'double',
                                'boolean',
                                'datetime',
                                'blob']:
                shape_cls = Shape
            else:
                raise InvalidShapeError("Unknown shape type: %s" % shape_type)
        except KeyError:
            raise InvalidShapeError("Shape is missing required key 'type': %s"
                                    % (shape_data))
        return shape_cls(name, shape_data, self)

    def get_shape_by_name(self, name, shape_name):
        try:
            shape_data = self._definitions[shape_name]
        except KeyError:
            raise NoShapeFoundError(shape_name)
        return self.get_shape(name, shape_data)

    def resolve_shape_ref(self, name, shape_ref):
        return self.get_shape_by_name(name, shape_ref[len('#/definitions/'):])


class OperationModel(object):

    def __init__(self,
                 operation_data,
                 service_model,
                 name,
                 http_method,
                 request_uri):
        self._operation_data = operation_data
        self._service_model = service_model
        self._name = name
        self.http = {
            'method': http_method,
            'requestUri': request_uri,
            }

    @CachedProperty
    def name(self):
        return self._name

    @property
    def service_model(self):
        return self._service_model

    @CachedProperty
    def summary(self):
        return self._operation_data["summary"]

    @CachedProperty
    def documentation(self):
        return self._operation_data["description"]

    @CachedProperty
    def can_paginate(self):
        # A valid paginating operation always has x-paging-default-max-items
        return "x-paging-default-max-items" in self._operation_data

    @CachedProperty
    def paging_input_token(self):
        if self.can_paginate:
            for name, shape in self.input_shape.members.items():
                if shape.is_paging_input_token:
                    return name
        return None

    @CachedProperty
    def paging_output_token(self):
        if self.can_paginate:
            for name, shape in self.output_shape.members.items():
                if shape.is_paging_output_token:
                    return name
        return None

    @CachedProperty
    def paging_result(self):
        if self.can_paginate:
            for name, shape in self.output_shape.members.items():
                if shape.is_paging_result:
                    return name
        return None

    @CachedProperty
    def paging_page_size(self):
        if self.can_paginate:
            for name, shape in self.input_shape.members.items():
                if shape.is_page_size:
                    return name
        return None

    @CachedProperty
    def paging_default_max_items(self):
        if self.can_paginate:
            return int(self._operation_data.get("x-paging-default-max-items"))
        return None

    @CachedProperty
    def form_factors(self):
        if "x-form-factors" in self._operation_data:
            return self._operation_data['x-form-factors'].split(',')
        return None

    @CachedProperty
    def extensions(self):
        if "x-extensions" in self._operation_data:
            return self._operation_data['x-extensions'].split(',')
        return None

    @CachedProperty
    def input_shape(self):
        return self._service_model.resolve_shape_ref(
            "input", self._operation_data['parameters'][0]['schema'][REF_KEY])

    @CachedProperty
    def output_shape(self):
        if '200' in self._operation_data['responses']:
            return self._service_model.resolve_shape_ref(
                "output", self._operation_data['responses']['200']['schema'][REF_KEY])
        else:
            return self._service_model.resolve_shape_ref(
                "output", self._operation_data['responses'][200]['schema'][REF_KEY])


class ServiceModel(object):

    def __init__(self, service_data, service_name):
        self._service_data = service_data
        self._service_name = service_name
        self._instance_cache = {}
        self._shape_resolver = ShapeResolver(service_data.get('definitions', {}))

    @CachedProperty
    def service_name(self):
        return self._service_name

    @CachedProperty
    def service_version(self):
        return self._service_data["info"]["version"]

    @CachedProperty
    def service_title(self):
        return self._service_data["info"]["title"]

    @CachedProperty
    def documentation(self):
        return self._service_data["info"]["description"]

    @instance_cache
    def operation_model(self, operation_name):
        for request_uri in self._service_data["paths"]:
            for http_method in self._service_data["paths"][request_uri]:
                operation_data = self._service_data["paths"][request_uri][http_method]
                if operation_data["operationId"] == operation_name:
                    return OperationModel(operation_data,
                                          self,
                                          operation_name,
                                          http_method,
                                          request_uri)
        raise OperationNotFoundError(operation_name)

    @CachedProperty
    def operation_names(self):
        operation_names = []
        for path in self._service_data["paths"]:
            for operation in self._service_data["paths"][path]:
                operation_names.append(
                    self._service_data["paths"][path][operation]["operationId"])
        return operation_names

    def resolve_shape_ref(self, name, shape_ref):
        return self._shape_resolver.resolve_shape_ref(name, shape_ref)

    @CachedProperty
    def endpoint_name(self):
        return self._service_data["x-endpoint-name"]

    @CachedProperty
    def endpoint_prefix(self):
        return self._service_data.get("x-endpoint-prefix", 'api')

    @CachedProperty
    def products(self):
        return self._service_data.get("x-products", "ALTUS").split(',')

    @CachedProperty
    def form_factors(self):
        return self._service_data.get("x-form-factors", "public").split(',')


class DenormalizedStructureBuilder(object):
    """Build a StructureShape from a denormalized model.

    This is a convenience builder class that makes it easy to construct
    ``StructureShape``s based on a denormalized model.

    It will handle the details of creating unique shape names and creating
    the appropriate shape map needed by the ``StructureShape`` class.

    Example usage::

        builder = DenormalizedStructureBuilder()
        shape = builder.with_members({
            'A': {
                'type': 'object',
                'properties': {
                    'B': {
                        'type': 'object',
                        'properties': {
                            'C': {
                                'type': 'string',
                            }
                        }
                    }
                }
            }
        }).build_model()
        # ``shape`` is now an instance of cdpcli.model.StructureShape
    """
    def __init__(self, name=None):
        self.members = OrderedDict()
        self._name_generator = ShapeNameGenerator()
        if name is None:
            self.name = self._name_generator.new_shape_name('object')

    def with_members(self, members):
        self._members = members
        return self

    def build_model(self):
        shapes = OrderedDict()
        denormalized = {
            'type': OBJECT_TYPE,
            'properties': self._members,
        }
        self._build_model(denormalized, shapes, self.name)
        resolver = ShapeResolver(shapes)
        return ObjectShape(name=self.name,
                           shape_data=shapes[self.name],
                           shape_resolver=resolver)

    def _build_model(self, model, shapes, shape_name):
        shape_type = _get_shape_type(model)
        if shape_type == OBJECT_TYPE:
            shapes[shape_name] = self._build_object(model, shapes)
        elif shape_type == MAP_TYPE:
            shapes[shape_name] = self._build_map(model, shapes)
        elif shape_type == LIST_TYPE:
            shapes[shape_name] = self._build_array(model, shapes)
        elif shape_type in ['string',
                            'integer',
                            'float',
                            'double',
                            'boolean',
                            'datetime',
                            'blob']:
            shapes[shape_name] = self._build_scalar(model)
        else:
            raise InvalidShapeError("Unknown shape type: %s" % model['type'])

    def _build_object(self, model, shapes):
        members = OrderedDict()
        shape = self._build_initial_shape(model)
        shape['properties'] = members

        for name, member_model in model['properties'].items():
            member_shape_name = self._get_shape_name(member_model)
            members[name] = {REF_KEY: REF_NAME_PREFIX + member_shape_name}
            self._build_model(member_model, shapes, member_shape_name)
        return shape

    def _build_array(self, model, shapes):
        member_shape_name = self._get_shape_name(model['items'])
        shape = self._build_initial_shape(model)
        shape['items'] = {REF_KEY: REF_NAME_PREFIX + member_shape_name}
        self._build_model(model['items'], shapes, member_shape_name)
        return shape

    def _build_map(self, model, shapes):
        value_shape_name = self._get_shape_name(model[ADDITIONAL_PROPERTIES])
        shape = self._build_initial_shape(model)
        shape[ADDITIONAL_PROPERTIES] = {REF_KEY: REF_NAME_PREFIX + value_shape_name}
        self._build_model(model[ADDITIONAL_PROPERTIES], shapes, value_shape_name)
        return shape

    def _build_initial_shape(self, model):
        shape = {
            'type': model['type'],
        }
        if 'documentation' in model:
            shape['documentation'] = model['documentation']
        if 'enum' in model:
            shape['enum'] = model['enum']
        if 'x-supported-options' in model:
            shape['x-supported-options'] = model['x-supported-options']
        if 'x-no-paramfile' in model:
            shape['x-no-paramfile'] = model['x-no-paramfile']
        return shape

    def _build_scalar(self, model):
        return self._build_initial_shape(model)

    def _get_shape_name(self, model):
        if 'shape_name' in model:
            return model['shape_name']
        else:
            return self._name_generator.new_shape_name(model['type'])


class ShapeNameGenerator(object):
    """Generate unique shape names for a type.

    This class can be used in conjunction with the DenormalizedStructureBuilder
    to generate unique shape names for a given type.

    """
    def __init__(self):
        self._name_cache = defaultdict(int)

    def new_shape_name(self, type_name):
        """Generate a unique shape name.

        This method will guarantee a unique shape name each time it is
        called with the same type.

        ::

            >>> s = ShapeNameGenerator()
            >>> s.new_shape_name('object')
            'ObjectType1'
            >>> s.new_shape_name('object')
            'ObjectType2'
            >>> s.new_shape_name('array')
            'ListType1'
            >>> s.new_shape_name('array')
            'ListType2'


        :type type_name: string
        :param type_name: The type name (object, list, map, string, etc.)

        :rtype: string
        :return: A unique shape name for the given type

        """
        self._name_cache[type_name] += 1
        current_index = self._name_cache[type_name]
        return '%sType%s' % (type_name.capitalize(),
                             current_index)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from base64 import urlsafe_b64decode
import datetime

import cdpcli.auth
from cdpcli.cdprequest import CdpRequest
from cdpcli.compat import HTTPHeaders
from cdpcli.compat import json
from cdpcli.compat import urlsplit
import cdpcli.credentials
from cdpcli.exceptions import NoCredentialsError
import mock
from tests import unittest
from tests.unit import ECDSA_KEY, ED25519_KEY, RSA_KEY


EXPECTED_RSA_SIG = \
    'VLlOczaMiHdAHfW7-0axYWAxpFPqHR2sR22XRh98AlVTBjj8QJTModpzNUQxb1N0F94pMP6U' \
    'BI-flm-rl3vHJaRBfcWbaDglD02YcuqD87CmOIpZ6Z3TUTbkOcxTsMSkgOaPqQkO1p49WRl3' \
    'P_v3Q9z5y6Mh7ZDbQeonQcagKhoIYQnCXYrEmAAHhTGwuxanuAsPu2y8svUBKNd9fXZ7stQ0' \
    'Pom2J2aQZnegBM6I_QJICP7ZEd0Roga0AcGoL1OsZo_fANkUV9eUvtw8CfTw11G2c1YS__pq' \
    'PVuW4iPSYONoUN5NrL6x3RtGOea0Xo__9B5ki0_TLsYMyhF37it6qA=='

EXPECTED_ED25519_SIG = \
    'et_Ueu_w3QuQbqvDdwy9aT8HLzsXBdJfRRRHe4fEK_RZ-qR-xvM35XG8J8q-YMz70GunK82JoSt5ztz0lAuCBg=='  # noqa

EXPECTED_ECDSA_SIG = \
    'MIGIAkIBPOgngVp4-uyWzLTCKcsCBeYezoedjN-Tj2KI78aIn6huJEJwG-6ZiVcEyO-mI1wN' \
    'tbVu7H-iQDMPFKyBdpezjxYCQgEY32q1WTQUPHj_JNYN33QhCMPjJHIaBcHTdMp0SO3iVzLD' \
    'Xi1MmHoMGsIYnUq3Q8Y1Z5JiyFr0NO6XzfnrF_OYCQ=='


class BaseTestWithFixedDate(unittest.TestCase):
    def setUp(self):
        self.datetime_patch = mock.patch('cdpcli.auth.datetime')
        self.datetime_mock = self.datetime_patch.start()
        self.fixed_date = datetime.datetime(2014, 3, 10, 17, 2, 55, 0)
        self.datetime_mock.datetime.utcnow.return_value = self.fixed_date
        self.datetime_mock.datetime.strptime.return_value = self.fixed_date

    def tearDown(self):
        self.datetime_patch.stop()


class TestBaseSigner(unittest.TestCase):
    def test_base_signer_raises_error(self):
        base_signer = cdpcli.auth.BaseSigner()
        with self.assertRaises(NotImplementedError):
            base_signer.add_auth("pass someting")


class TestRSAV1(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        access_key_id = 'ABCD-EFGH-IJKL-MNOP-QRST'
        self.credentials = cdpcli.credentials.Credentials(
            access_key_id=access_key_id,
            private_key=RSA_KEY,
            method='test')
        self.rsav1 = cdpcli.auth.RSAv1Auth(self.credentials)
        self.date_mock = mock.patch('cdpcli.auth.formatdate')
        self.formatdate = self.date_mock.start()
        self.formatdate.return_value = 'Thu, 17 Nov 2005 18:49:58 GMT'

    def tearDown(self):
        self.date_mock.stop()

    def test_put(self):
        http_headers = HTTPHeaders.from_dict({})
        split = urlsplit('/foo/bar')
        cs = self.rsav1._canonical_string('PUT', split, http_headers)
        expected_canonical = "PUT\n\nThu, 17 Nov 2005 18:49:58 GMT\n/foo/bar\nrsav1"
        self.assertEqual(expected_canonical, cs)
        sig = self.rsav1._get_signature('PUT', split, HTTPHeaders.from_dict({}))
        self.assertEqual(EXPECTED_RSA_SIG, sig)

    def test_duplicate_date(self):
        pairs = [('x-altus-date', 'Thu, 17 Nov 2015 18:49:58 GMT'),
                 ('X-Altus-Magic', 'abracadabra')]
        http_headers = HTTPHeaders.from_pairs(pairs)
        split = urlsplit('/foo/bar')
        with self.assertRaises(Exception):
            self.rsav1.get_signature('PUT', split, http_headers)

    def test_duplicate_auth_header(self):
        request = CdpRequest()
        request.headers = HTTPHeaders.from_dict({'x-altus-auth': 'signature'})
        request.method = 'PUT'
        request.url = 'https://altus.cloudera.com/service/op'
        with self.assertRaises(Exception):
            self.rsav1._inject_signature(request, 'new_signature')

    def test_resign_uses_most_recent_date(self):
        dates = [
            'Thu, 17 Nov 2005 18:49:58 GMT',
            'Thu, 17 Nov 2014 20:00:00 GMT',
        ]
        self.formatdate.side_effect = dates

        request = CdpRequest()
        request.headers['Content-Type'] = 'text/html'
        request.method = 'PUT'
        request.url = 'https://altus.cloudera.com/service/op'

        self.rsav1.add_auth(request)
        original_date = request.headers['x-altus-date']

        del request.headers['x-altus-date']
        del request.headers['x-altus-auth']
        self.rsav1.add_auth(request)
        modified_date = request.headers['x-altus-date']

        # Each time we sign a request, we make another call to formatdate()
        # so we should have a different date header each time.
        self.assertEqual(original_date, dates[0])
        self.assertEqual(modified_date, dates[1])

    def test_no_credentials_raises_error(self):
        rsav1 = cdpcli.auth.RSAv1Auth(None)
        with self.assertRaises(NoCredentialsError):
            rsav1.add_auth("pass someting")

    def test_auth_header_string(self):
        http_headers = HTTPHeaders.from_dict({})
        split = urlsplit('/foo/bar')
        sig = self.rsav1._get_signature('PUT', split, http_headers)
        self.assertEqual(EXPECTED_RSA_SIG, sig)

        auth_header_string = self.rsav1._get_signature_header(sig)
        expected_metadata = 'eyJhY2Nlc3Nfa2V5X2lkIjogIkFCQ0QtRUZHSC1JSktMLU1O' \
                            'T1AtUVJTVCIsICJhdXRoX21ldGhvZCI6ICJyc2F2MSJ9'
        metadata, sig = auth_header_string.split(".")
        self.assertEqual(expected_metadata, metadata)
        self.assertEqual(EXPECTED_RSA_SIG, sig)

        json_metadata = json.loads(
            urlsafe_b64decode(metadata.encode('utf-8')).decode('utf-8'))
        self.assertEqual(self.credentials.access_key_id,
                         json_metadata['access_key_id'])
        self.assertEqual("rsav1",
                         json_metadata['auth_method'])


class TestED25519V1(unittest.TestCase):
    """
    We're not retesting aspects that are identical to what RSA tests cover
    """

    def setUp(self):
        access_key_id = 'ABCD-EFGH-IJKL-MNOP-QRST'
        self.credentials = cdpcli.credentials.Credentials(
            access_key_id=access_key_id,
            private_key=ED25519_KEY,
            method='test')
        self.ed25519v1 = cdpcli.auth.Ed25519v1Auth(self.credentials)
        self.date_mock = mock.patch('cdpcli.auth.formatdate')
        self.formatdate = self.date_mock.start()
        self.formatdate.return_value = 'Thu, 17 Nov 2005 18:49:58 GMT'

    def tearDown(self):
        self.date_mock.stop()

    def test_put(self):
        http_headers = HTTPHeaders.from_dict({})
        split = urlsplit('/foo/bar')
        cs = self.ed25519v1._canonical_string('PUT', split, http_headers)
        expected_canonical = "PUT\n\nThu, 17 Nov 2005 18:49:58 GMT\n/foo/bar\ned25519v1"
        self.assertEqual(expected_canonical, cs)
        sig = self.ed25519v1._get_signature('PUT', split, HTTPHeaders.from_dict({}))
        self.assertEqual(EXPECTED_ED25519_SIG, sig)

    def test_no_credentials_raises_error(self):
        ed25519v1 = cdpcli.auth.Ed25519v1Auth(None)
        with self.assertRaises(NoCredentialsError):
            ed25519v1.add_auth("pass someting")

    def test_auth_header_string(self):
        http_headers = HTTPHeaders.from_dict({})
        split = urlsplit('/foo/bar')
        sig = self.ed25519v1._get_signature('PUT', split, http_headers)
        self.assertEqual(EXPECTED_ED25519_SIG, sig)

        auth_header_string = self.ed25519v1._get_signature_header(sig)
        expected_metadata = 'eyJhY2Nlc3Nfa2V5X2lkIjogIkFCQ0QtRUZHSC1JSktMLU1OT1' \
                            'AtUVJTVCIsICJhdXRoX21ldGhvZCI6ICJlZDI1NTE5djEifQ=='
        metadata, sig = auth_header_string.split(".")
        self.assertEqual(expected_metadata, metadata)
        self.assertEqual(EXPECTED_ED25519_SIG, sig)

        json_metadata = json.loads(
            urlsafe_b64decode(metadata.encode('utf-8')).decode('utf-8'))
        self.assertEqual(self.credentials.access_key_id,
                         json_metadata['access_key_id'])
        self.assertEqual("ed25519v1",
                         json_metadata['auth_method'])


class TestECDSAV1(unittest.TestCase):
    """
    We're not retesting aspects that are identical to what RSA tests cover
    """

    def setUp(self):
        access_key_id = 'ABCD-EFGH-IJKL-MNOP-QRST'
        self.credentials = cdpcli.credentials.Credentials(
            access_key_id=access_key_id,
            private_key=ECDSA_KEY,
            method='test')
        self.ecdsav1 = cdpcli.auth.ECDSAv1Auth(self.credentials)
        self.date_mock = mock.patch('cdpcli.auth.formatdate')
        self.formatdate = self.date_mock.start()
        self.formatdate.return_value = 'Thu, 17 Nov 2005 18:49:58 GMT'

    def tearDown(self):
        self.date_mock.stop()

    def test_no_credentials_raises_error(self):
        ecdsav1 = cdpcli.auth.ECDSAv1Auth(None)
        with self.assertRaises(NoCredentialsError):
            ecdsav1.add_auth("pass someting")

    def test_put(self):
        http_headers = HTTPHeaders.from_dict({})
        split = urlsplit('/foo/bar')
        cs = self.ecdsav1._canonical_string('PUT', split, http_headers)
        expected_canonical = "PUT\n\nThu, 17 Nov 2005 18:49:58 GMT\n/foo/bar\necdsav1"
        self.assertEqual(expected_canonical, cs)
        sig = self.ecdsav1._get_signature('PUT', split, HTTPHeaders.from_dict({}))
        self.assertEqual(EXPECTED_ECDSA_SIG, sig)

    def test_auth_header_string(self):
        http_headers = HTTPHeaders.from_dict({})
        split = urlsplit('/foo/bar')
        sig = self.ecdsav1._get_signature('PUT', split, http_headers)
        self.assertEqual(EXPECTED_ECDSA_SIG, sig)

        auth_header_string = self.ecdsav1._get_signature_header(sig)
        expected_metadata = 'eyJhY2Nlc3Nfa2V5X2lkIjogIkFCQ0QtRUZHSC1JSktMLU1O' \
                            'T1AtUVJTVCIsICJhdXRoX21ldGhvZCI6ICJlY2RzYXYxIn0='
        metadata, sig = auth_header_string.split(".")
        self.assertEqual(expected_metadata, metadata)
        self.assertEqual(EXPECTED_ECDSA_SIG, sig)

        json_metadata = json.loads(
            urlsafe_b64decode(metadata.encode('utf-8')).decode('utf-8'))
        self.assertEqual(self.credentials.access_key_id, json_metadata['access_key_id'])
        self.assertEqual("ecdsav1", json_metadata['auth_method'])


class TestAccessToken(unittest.TestCase):
    def test_no_credentials_raises_error(self):
        access_token_auth = cdpcli.auth.AccessTokenAuth(None)
        with self.assertRaises(NoCredentialsError):
            access_token_auth.add_auth("pass something")

    def test_auth_header_string(self):
        access_token = 'Bearer ABC.DEF'
        credentials = cdpcli.credentials.Credentials(
            access_token=access_token,
            method='test')
        access_token_auth = cdpcli.auth.AccessTokenAuth(credentials)

        request = CdpRequest()
        request.headers['Content-Type'] = 'text/html'
        request.method = 'PUT'
        request.url = 'https://altus.cloudera.com/service/op'

        access_token_auth.add_auth(request)
        self.assertEqual(request.headers['Authorization'], access_token)

    def test_is_access_token(self):
        self.assertFalse(cdpcli.auth.AccessTokenAuth.is_access_token(None))
        self.assertFalse(cdpcli.auth.AccessTokenAuth.is_access_token(''))
        self.assertTrue(cdpcli.auth.AccessTokenAuth.is_access_token('Bearer A.B'))



# Licensed to Cloudera, Inc. under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  Cloudera, Inc. licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from cdpcli.clidriver import CLIOperationCaller, ServiceOperation
from cdpcli.exceptions import ClientError, DfExtensionError
from cdpcli.extensions.df import get_expanded_file_path, upload_workload_asset
from cdpcli.extensions.workload import set_workload_access_token
from cdpcli.model import ObjectShape, OperationModel, ShapeResolver
from cdpcli.utils import CachedProperty

LOG = logging.getLogger('cdpcli.extensions.df.createdeployment')
INITIAL_CONFIGURATION_VERSION = 0
INITIAL_ASSET_VERSION = '0'

SERVICE_NAME = 'df'
OPERATION_NAME = 'createDeployment'
OPERATION_CLI_NAME = 'create-deployment'
OPERATION_SUMMARY = 'Initiate and create deployment on workload'
OPERATION_DESCRIPTION = """
    Initiate deployment on the control plane, upload referenced when specified,
    and create the deployment on workload. This operation is supported for the CLI only.'
    """
OPERATION_DATA = {
    'summary': OPERATION_SUMMARY,
    'description': OPERATION_DESCRIPTION,
    'operationId': OPERATION_NAME,
}
OPERATION_SHAPES = {
    'CreateDeploymentRequest': {
        'type': 'object',
        'description': 'Request object for creating a deployment.',
        'required': ['serviceCrn', 'flowVersionCrn', 'deploymentName'],
        'properties': {
            'serviceCrn': {
                'type': 'string',
                'description': 'CRN for the service.'
            },
            'flowVersionCrn': {
                'type': 'string',
                'description': 'CRN for the flow definition version.'
            },
            'deploymentName': {
                'type': 'string',
                'description': 'Unique name for the deployment.'
            },
            'clusterSizeName': {
                'type': 'string',
                'description': 'Size for the cluster. The default is EXTRA_SMALL.',
                'enum': [
                    'EXTRA_SMALL',
                    'SMALL',
                    'MEDIUM',
                    'LARGE'
                ],
            },
            'staticNodeCount': {
                'type': 'integer',
                'description': 'The static number of nodes provisioned. The default is 1.'
            },
            'autoScalingEnabled': {
                'type': 'boolean',
                'description': 'Automatic node scaling. The default is disabled.'
            },
            'autoScaleMinNodes': {
                'type': 'integer',
                'description': 'The minimum number of nodes for automatic scaling.'
            },
            'autoScaleMaxNodes': {
                'type': 'integer',
                'description': 'The maximum number of nodes for automatic scaling.'
            },
            'cfmNifiVersion': {
                'type': 'string',
                'description': 'The CFM NiFi version. Defaults to the latest version.'
            },
            'autoStartFlow': {
                'type': 'boolean',
                'description': 'Automatically start the flow.'
            },
            'parameterGroups': {
                'type': 'array',
                'description': 'Parameter groups with each requiring a value or assets.',
                'items': {
                    '$ref': '#/definitions/DeploymentFlowParameterGroup'
                }
            },
            'kpis': {
                'type': 'array',
                'description': 'Key Performance Indicators with associated alerts.',
                'items': {
                    '$ref': '#/definitions/DeploymentKeyPerformanceIndicator'
                }
            },
            'customNarConfiguration': {
                'type': 'object',
                'description': 'Custom NAR configuration properties',
                'required': [
                    'username',
                    'password',
                    'storageLocation',
                    'configurationVersion'
                ],
                'properties': {
                    'username': {
                        'type': 'string',
                        'description': 'Username for access to NAR storage location'
                    },
                    'password': {
                        'type': 'string',
                        'description': 'Password for access to NAR storage location',
                        'x-sensitive': 'true'
                    },
                    'storageLocation': {
                        'type': 'string',
                        'description': 'Storage location containing custom NAR files',
                        'x-no-paramfile': 'true'
                    },
                    'configurationVersion': {
                        'type': 'integer',
                        'description': 'Custom configuration version number'
                    }
                }
            },
            'inboundHostname': {
                'type': 'string',
                'description':
                    'The FQDN of inbound host or just the prefix part of the hostname'
            },
            'listenComponents': {
                'type': 'array',
                'description': 'Listen components port and protocol data',
                'items': {
                    '$ref': '#/definitions/ListenComponent'
                }
            }
        }
    },
    'CreateDeploymentResponse': {
        'type': 'object',
        'description': 'Response for Create Deployment command.',
        'properties': {
            'crn': {
                'type': 'string',
                'description': 'CRN for the created deployment.'
            }
        }
    },
    'DeploymentFlowParameterGroup': {
        'type': 'object',
        'description': 'Parameter groups for NiFi flow deployment.',
        'required': ['name', 'parameters'],
        'properties': {
            'name': {
                'type': 'string',
                'description': 'Name for parameter group.'
            },
            'parameters': {
                'type': 'array',
                'description': 'Parameters for the parameter group.',
                'items': {
                    '$ref': '#/definitions/DeploymentFlowParameter'
                }
            },
        }
    },
    'DeploymentFlowParameter': {
        'type': 'object',
        'description': 'Parameter object for the NiFi flow deployment.',
        'required': ['name'],
        'properties': {
            'name': {
                'type': 'string',
                'description': 'Name for the parameter.'
            },
            'value': {
                'type': 'string',
                'description': 'Value for the named parameter.',
                'x-no-paramfile': 'true'
            },
            'assetReferences': {
                'type': 'array',
                'description': 'Asset references for the named parameter.',
                'items': {
                    'type': 'string'
                }
            }
        }
    },
    'DeploymentKeyPerformanceIndicator': {
        'type': 'object',
        'description': 'Key Performance Indicators for the deployment.',
        'required': ['metricId'],
        'properties': {
            'metricId': {
                'type': 'string',
                'description': 'Unique identifier for the metric object.'
            },
            'componentId': {
                'type': 'string',
                'description': 'Identifier for the NiFi component.'
            },
            'alert': {
                '$ref': '#/definitions/DeploymentAlert'
            }
        }
    },
    'DeploymentAlert': {
        'type': 'object',
        'properties': {
            'thresholdMoreThan': {
                '$ref': '#/definitions/DeploymentAlertThreshold',
                'description': 'The threshold above which alerts should be triggered.'
            },
            'thresholdLessThan': {
                '$ref': '#/definitions/DeploymentAlertThreshold',
                'description': 'The threshold below which alerts should be triggered.'
            },
            'frequencyTolerance': {
                '$ref': '#/definitions/DeploymentFrequencyTolerance'
            }
        }
    },
    'DeploymentAlertThreshold': {
        'type': 'object',
        'properties': {
            'unitId': {
                'type': 'string',
                'description': 'The unit identifier for the alert threshold.'
            },
            'value': {
                'type': 'double',
                'description': 'The numeric value for the alert threshold.'
            }
        }
    },
    'DeploymentFrequencyTolerance': {
        'type': 'object',
        'description': 'The frequency tolerance for the Key Performance Indicator.',
        'properties': {
            'value': {
                'type': 'double',
                'description': 'The amount of time before generating an alert.'
            },
            'unit': {
                'type': 'object',
                'description': 'The time unit for associated value number.',
                'properties': {
                    'id': {
                        'type': 'string',
                        'enum': [
                            'SECONDS',
                            'MINUTES',
                            'HOURS',
                            'DAYS'
                        ]
                    }
                }
            }
        }
    },
    'ListenComponent': {
        'type': 'object',
        'required': ['protocol', 'port'],
        'properties': {
            'protocol': {
                'type': 'string',
                'description': 'Inbound protocol',
                'enum': ['TCP', 'UDP']
            },
            'port': {
                'type': 'string',
                'description': 'Inbound port'
            }
        },
        'description': 'Provides subset of metadata of a Listen* component'
    }
}


class CreateDeployment(ServiceOperation):

    def __init__(self, clidriver, service_model):
        super(CreateDeployment, self).__init__(
            clidriver=clidriver,
            name=OPERATION_CLI_NAME,
            parent_name=SERVICE_NAME,
            operation_model=CreateDeploymentOperationModel(service_model),
            operation_caller=CreateDeploymentOperationCaller())


class CreateDeploymentOperationModel(OperationModel):

    def __init__(self, service_model):
        super(CreateDeploymentOperationModel, self).__init__(
            operation_data=OPERATION_DATA,
            service_model=service_model,
            name=OPERATION_NAME,
            http_method=None,
            request_uri=None)

    @CachedProperty
    def input_shape(self):
        resolver = ShapeResolver(OPERATION_SHAPES)
        return ObjectShape(name='input',
                           shape_data=OPERATION_SHAPES['CreateDeploymentRequest'],
                           shape_resolver=resolver)

    @CachedProperty
    def output_shape(self):
        resolver = ShapeResolver(OPERATION_SHAPES)
        return ObjectShape(name='output',
                           shape_data=OPERATION_SHAPES['CreateDeploymentResponse'],
                           shape_resolver=resolver)


class CreateDeploymentOperationCaller(CLIOperationCaller):

    def invoke(self,
               client_creator,
               operation_model,
               parameters,
               parsed_args,
               parsed_globals):
        df_client = client_creator('df')

        service_crn = parameters.get('serviceCrn', None)
        environment_crn = self._get_environment_crn(df_client, service_crn)
        flow_version_crn = parameters.get('flowVersionCrn', None)
        deployment_request_crn = self._initiate_deployment(
                df_client, service_crn, flow_version_crn)

        iam_client = client_creator('iam')
        set_workload_access_token(iam_client, parsed_globals, SERVICE_NAME.upper(),
                                  environment_crn)

        df_workload_client = client_creator('dfworkload')
        self._upload_assets(df_workload_client, deployment_request_crn, parameters)
        response = self._create_deployment(
                df_workload_client, deployment_request_crn, environment_crn, parameters)
        self._display_response(operation_model.name, response, parsed_globals)

    def _create_deployment(self,
                           df_workload_client,
                           deployment_request_crn,
                           environment_crn,
                           parameters):
        """
        Create Deployment on Workload using initiated Deployment Request CRN
        """
        deployment_configuration = self._get_deployment_configuration(
                deployment_request_crn, parameters)

        nar_configuration_crn = self._process_custom_nar_configuration(
                df_workload_client, environment_crn, parameters)
        if nar_configuration_crn:
            deployment_configuration['customNarConfigurationCrn'] = nar_configuration_crn

        deployment_configuration['environmentCrn'] = environment_crn
        LOG.debug('Create Deployment Parameters %s', deployment_configuration)
        http, response = df_workload_client.make_api_call(
                'createDeployment', deployment_configuration)
        deployment = response.get('deployment', {})
        deployment_crn = deployment.get('crn', None)
        create_response = {
            'deploymentCrn': deployment_crn
        }
        return create_response

    def _initiate_deployment(self,
                             df_client,
                             service_crn,
                             flow_version_crn):
        """
        Initiate Deployment on Control Plane using Service CRN and Flow Version CRN
        """
        deployment_parameters = {
            'serviceCrn': service_crn,
            'flowVersionCrn': flow_version_crn
        }
        http, response = df_client.make_api_call(
                'initiateDeployment', deployment_parameters)
        request_crn = response.get('deploymentRequestCrn', None)
        url = response.get('dfxLocalUrl', None)
        LOG.debug('Initiated Deployment Request CRN [%s] URL [%s]', request_crn, url)
        return request_crn

    def _get_environment_crn(self,
                             df_client,
                             service_crn):
        """
        Get Environment CRN using Service CRN
        """
        http, response = df_client.make_api_call(
                'listDeployableServicesForNewDeployments', {})
        services = response.get('services', [])

        for service in services:
            crn = service.get('crn', None)
            if service_crn == crn:
                environment_crn = service.get('environmentCrn', None)
                LOG.debug('Found Environment CRN [%s]', environment_crn)
                return environment_crn
        raise DfExtensionError(err_msg='Environment CRN not found for Service CRN',
                               service_name='df',
                               operation_name='listDeployableServicesForNewDeployments')

    def _process_custom_nar_configuration(self,
                                          df_workload_client,
                                          environment_crn,
                                          parameters):
        """
        Process Custom NAR Configuration and return Custom NAR Configuration CRN
        """
        custom_nar_configuration = parameters.get('customNarConfiguration', None)
        if custom_nar_configuration:
            custom_nar_configuration['environmentCrn'] = environment_crn

            default_parameters = {
                'environmentCrn': environment_crn
            }

            try:
                default_http, default_configuration = df_workload_client.make_api_call(
                        'getDefaultCustomNarConfiguration', default_parameters)
                custom_nar_configuration['crn'] = default_configuration.get('crn', None)
                default_version = default_configuration.get('configurationVersion', None)
                custom_nar_configuration['configurationVersion'] = default_version

                http, response = df_workload_client.make_api_call(
                        'updateCustomNarConfiguration', custom_nar_configuration)
                crn = response.get('crn', None)
                LOG.debug('Updated Custom NAR Configuration CRN [%s]', crn)
                return crn
            except ClientError as e:
                if e.http_status_code == 404:
                    http, response = df_workload_client.make_api_call(
                            'createCustomNarConfiguration', custom_nar_configuration)
                    crn = response.get('crn', None)
                    LOG.debug('Created Custom NAR Configuration CRN [%s]', crn)
                    return crn
                else:
                    raise
        else:
            return None

    def _get_deployment_configuration(self,
                                      deployment_request_crn,
                                      parameters):
        """
        Get Deployment Configuration request based on command parameters
        """
        deployment_configuration = {
            'name': parameters.get('deploymentName', None),
            'deploymentRequestCrn': deployment_request_crn,
            'configurationVersion': INITIAL_CONFIGURATION_VERSION,
            'clusterSizeName': parameters.get('clusterSizeName', 'EXTRA_SMALL'),
            'staticNodeCount': parameters.get('staticNodeCount', 1)
        }

        inboundHostname = parameters.get('inboundHostname', None)
        if inboundHostname is not None:
            deployment_configuration['inboundHostname'] = inboundHostname
            deployment_configuration['listenComponents'] = \
                parameters.get('listenComponents', None)

        autoScalingEnabled = parameters.get('autoScalingEnabled', None)
        if autoScalingEnabled is not None:
            deployment_configuration['autoScalingEnabled'] = autoScalingEnabled
        if autoScalingEnabled:
            del deployment_configuration['staticNodeCount']

            autoScaleMinNodes = parameters.get('autoScaleMinNodes', None)
            if autoScaleMinNodes:
                deployment_configuration['autoScaleMinNodes'] = autoScaleMinNodes
            autoScaleMaxNodes = parameters.get('autoScaleMaxNodes', None)
            if autoScaleMaxNodes:
                deployment_configuration['autoScaleMaxNodes'] = autoScaleMaxNodes

        autoStartFlow = parameters.get('autoStartFlow', None)
        if autoStartFlow is not None:
            deployment_configuration['autoStartFlow'] = autoStartFlow

        cfmNifiVersion = parameters.get('cfmNifiVersion', None)
        if cfmNifiVersion:
            deployment_configuration['cfmNifiVersion'] = cfmNifiVersion

        parameterGroups = parameters.get('parameterGroups', None)
        if parameterGroups:
            deployment_configuration['parameterGroups'] = parameterGroups
        kpis = parameters.get('kpis', None)
        if kpis:
            deployment_configuration['kpis'] = self._process_kpis(kpis)

        return deployment_configuration

    def _process_kpis(self, kpis):
        """
        Process Key Performance Indicators and set required unit properties
        """
        for kpi in kpis:
            alert = kpi.get('alert', None)
            if alert:
                frequency_tolerance = alert.get('frequencyTolerance', None)
                if frequency_tolerance:
                    unit = frequency_tolerance['unit']
                    id = unit['id']
                    unit['label'] = id.capitalize()
                    unit['abbreviation'] = id[:1].lower()
        return kpis

    def _upload_assets(self,
                       df_workload_client,
                       deployment_request_crn,
                       parameters):
        """
        Upload Assets associated with Deployment when Asset References found
        """
        parameter_groups = parameters.get('parameterGroups', None)
        if parameter_groups:
            deployment_name = parameters.get('deploymentName', None)
            for parameter_group in parameter_groups:
                parameter_group_name = parameter_group['name']
                parameters = parameter_group['parameters']
                for parameter in parameters:
                    asset_references = parameter.get('assetReferences', None)
                    if asset_references:
                        updated_asset_references = []
                        for asset_path in asset_references:
                            asset_params = {
                                'deploymentName': deployment_name,
                                'deploymentRequestCrn': deployment_request_crn,
                                'parameterGroup': parameter_group_name,
                                'parameterName': parameter.get('name', None),
                                'filePath': asset_path
                            }
                            upload_workload_asset(df_workload_client, asset_params)

                            file_path = get_expanded_file_path(asset_path)
                            path, name = os.path.split(file_path)
                            asset_reference = {
                                'name': name,
                                'path': path,
                                'version': INITIAL_ASSET_VERSION
                            }
                            updated_asset_references.append(asset_reference)

                        parameter['assetReferences'] = updated_asset_references



# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import datetime

from cdpcli import OBJECT_TYPE
from cdpcli import xform_name
from cdpcli.model import DenormalizedStructureBuilder
from cdpcli.model import ServiceModel
from cdpcli.utils import ArgumentGenerator
from cdpcli.utils import CachedProperty
from cdpcli.utils import datetime2timestamp
from cdpcli.utils import get_extension_registers
from cdpcli.utils import get_service_module_name
from cdpcli.utils import instance_cache
from cdpcli.utils import is_absolute_url
from cdpcli.utils import parse_timestamp
from cdpcli.utils import parse_to_aware_datetime
from dateutil.tz import tzoffset, tzutc
from tests import unittest


class TestExtension(unittest.TestCase):
    def test_get_extension_registers(self):
        reg_ext, reg_cmd = get_extension_registers('iam')
        self.assertIsNone(reg_ext)
        self.assertIsNone(reg_cmd)

        reg_ext, reg_cmd = get_extension_registers('redirect')
        self.assertIsNotNone(reg_ext)
        self.assertIsNone(reg_cmd)

        reg_ext, reg_cmd = get_extension_registers('workload')
        self.assertIsNotNone(reg_ext)
        self.assertIsNone(reg_cmd)

        reg_ext, reg_cmd = get_extension_registers('df')
        self.assertIsNotNone(reg_ext)
        self.assertIsNotNone(reg_cmd)

        reg_ext, reg_cmd = get_extension_registers('invalid')
        self.assertIsNone(reg_ext)
        self.assertIsNone(reg_cmd)


class TestUrl(unittest.TestCase):
    def test_is_absolute_url(self):
        self.assertTrue(is_absolute_url('https://host.com/api/v1/test'))
        self.assertFalse(is_absolute_url('/api/v1/test'))


class TestTransformName(unittest.TestCase):
    def test_upper_camel_case(self):
        self.assertEqual(xform_name('UpperCamelCase'), 'upper_camel_case')
        self.assertEqual(xform_name('UpperCamelCase', '-'), 'upper-camel-case')

    def test_lower_camel_case(self):
        self.assertEqual(xform_name('lowerCamelCase'), 'lower_camel_case')
        self.assertEqual(xform_name('lowerCamelCase', '-'), 'lower-camel-case')

    def test_consecutive_upper_case(self):
        self.assertEqual(xform_name('HTTPHeaders'), 'http_headers')
        self.assertEqual(xform_name('HTTPHeaders', '-'), 'http-headers')

    def test_consecutive_upper_case_middle_string(self):
        self.assertEqual(xform_name('MainHTTPHeaders'), 'main_http_headers')
        self.assertEqual(xform_name('MainHTTPHeaders', '-'), 'main-http-headers')

    def test_s3_prefix(self):
        self.assertEqual(xform_name('S3BucketName'), 's3_bucket_name')

    def test_already_snake_cased(self):
        self.assertEqual(xform_name('leave_alone'), 'leave_alone')
        self.assertEqual(xform_name('s3_bucket_name'), 's3_bucket_name')
        self.assertEqual(xform_name('bucket_s3_name'), 'bucket_s3_name')

    def test_special_case_ends_with_s(self):
        self.assertEqual(xform_name('GatewayCRNs', '-'), 'gateway-crns')


class TestParseTimestamps(unittest.TestCase):
    def test_parse_iso8601(self):
        self.assertEqual(
            parse_timestamp('1970-01-01T00:10:00.000Z'),
            datetime.datetime(1970, 1, 1, 0, 10, tzinfo=tzutc()))

    def test_parse_epoch(self):
        self.assertEqual(
            parse_timestamp(1222172800),
            datetime.datetime(2008, 9, 23, 12, 26, 40, tzinfo=tzutc()))

    def test_parse_epoch_zero_time(self):
        self.assertEqual(
            parse_timestamp(0),
            datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tzutc()))

    def test_parse_epoch_as_string(self):
        self.assertEqual(
            parse_timestamp('1222172800'),
            datetime.datetime(2008, 9, 23, 12, 26, 40, tzinfo=tzutc()))

    def test_parse_rfc822(self):
        self.assertEqual(
            parse_timestamp('Wed, 02 Oct 2002 13:00:00 GMT'),
            datetime.datetime(2002, 10, 2, 13, 0, tzinfo=tzutc()))

    def test_parse_invalid_timestamp(self):
        with self.assertRaises(ValueError):
            parse_timestamp('invalid date')


class TestDatetime2Timestamp(unittest.TestCase):
    def test_datetime2timestamp_naive(self):
        self.assertEqual(
            datetime2timestamp(datetime.datetime(1970, 1, 2)), 86400)

    def test_datetime2timestamp_aware(self):
        tzinfo = tzoffset("BRST", -10800)
        self.assertEqual(
            datetime2timestamp(datetime.datetime(1970, 1, 2, tzinfo=tzinfo)),
            97200)


class TestParseToUTCDatetime(unittest.TestCase):
    def test_handles_utc_time(self):
        original = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tzutc())
        self.assertEqual(parse_to_aware_datetime(original), original)

    def test_handles_other_timezone(self):
        tzinfo = tzoffset("BRST", -10800)
        original = datetime.datetime(2014, 1, 1, 0, 0, 0, tzinfo=tzinfo)
        self.assertEqual(parse_to_aware_datetime(original), original)

    def test_handles_naive_datetime(self):
        original = datetime.datetime(1970, 1, 1, 0, 0, 0)
        expected = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tzutc())
        self.assertEqual(parse_to_aware_datetime(original), expected)

    def test_handles_string_epoch(self):
        expected = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tzutc())
        self.assertEqual(parse_to_aware_datetime('0'), expected)

    def test_handles_int_epoch(self):
        expected = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tzutc())
        self.assertEqual(parse_to_aware_datetime(0), expected)

    def test_handles_full_iso_8601(self):
        expected = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tzutc())
        self.assertEqual(
            parse_to_aware_datetime('1970-01-01T00:00:00Z'),
            expected)

    def test_year_only_iso_8601(self):
        expected = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=tzutc())
        self.assertEqual(parse_to_aware_datetime('1970-01-01'), expected)


class TestCachedProperty(unittest.TestCase):
    def test_cached_property_same_value(self):
        class CacheMe(object):
            @CachedProperty
            def foo(self):
                return 'foo'

        c = CacheMe()
        self.assertEqual(c.foo, 'foo')
        self.assertEqual(c.foo, 'foo')

    def test_cached_property_only_called_once(self):
        # Note: you would normally never want to cache
        # a property that returns a new value each time,
        # but this is done to demonstrate the caching behavior.

        class NoIncrement(object):
            def __init__(self):
                self.counter = 0

            @CachedProperty
            def current_value(self):
                self.counter += 1
                return self.counter

        c = NoIncrement()
        self.assertEqual(c.current_value, 1)
        # If the property wasn't cached, the next value should be
        # be 2, but because it's cached, we know the value will be 1.
        self.assertEqual(c.current_value, 1)


class TestArgumentGenerator(unittest.TestCase):
    def setUp(self):
        self.arg_generator = ArgumentGenerator()

    def assert_skeleton_from_model_is(self, model, generated_skeleton):
        shape = DenormalizedStructureBuilder().with_members(
            model).build_model()
        actual = self.arg_generator.generate_skeleton(shape)
        self.assertEqual(actual, generated_skeleton)

    def test_generate_string(self):
        self.assert_skeleton_from_model_is(
            model={
                'A': {'type': 'string'}
            },
            generated_skeleton={
                'A': ''
            }
        )

    def test_generate_scalars(self):
        self.assert_skeleton_from_model_is(
            model={
                'A': {'type': 'string'},
                'B': {'type': 'integer'},
                'C': {'type': 'float'},
                'D': {'type': 'double'},
                'E': {'type': 'boolean'},
                'F': {'type': 'datetime'},
                'G': {'type': 'blob'}
            },
            generated_skeleton={
                'A': '',
                'B': 0,
                'C': 0.0,
                'D': 0.0,
                'E': True,
                'F': '2002-10-02T13:00:00+00:00',
                'G': ''
            }
        )

    def test_generate_nested_objects(self):
        self.assert_skeleton_from_model_is(
            model={
                'A': {
                    'type': OBJECT_TYPE,
                    'properties': {
                        'B': {'type': 'string'},
                    }
                }
            },
            generated_skeleton={
                'A': {'B': ''}
            }
        )

    def test_generate_scalar_list(self):
        self.assert_skeleton_from_model_is(
            model={
                'A': {
                    'type': 'array',
                    'items': {
                        'type': 'string'
                    }
                },
            },
            generated_skeleton={
                'A': [''],
            }
        )

    def test_generate_scalar_map(self):
        self.assert_skeleton_from_model_is(
            model={
                'A': {
                    'type': 'object',
                    'additionalProperties': {
                        'type': 'string'
                    }
                }
            },
            generated_skeleton={
                'A': {
                    'KeyName': '',
                }
            }
        )


class TestInstanceCache(unittest.TestCase):
    class DummyClass(object):
        def __init__(self, cache):
            self._instance_cache = cache

        @instance_cache
        def add(self, x, y):
            return x + y

        @instance_cache
        def sub(self, x, y):
            return x - y

    def setUp(self):
        self.cache = {}

    def test_cache_single_method_call(self):
        adder = self.DummyClass(self.cache)
        self.assertEqual(adder.add(2, 1), 3)
        # This should result in one entry in the cache.
        self.assertEqual(len(self.cache), 1)
        # When we call the method with the same args,
        # we should reuse the same entry in the cache.
        self.assertEqual(adder.add(2, 1), 3)
        self.assertEqual(len(self.cache), 1)

    def test_can_cache_multiple_methods(self):
        adder = self.DummyClass(self.cache)
        adder.add(2, 1)

        # A different method results in a new cache entry,
        # so now there should be two elements in the cache.
        self.assertEqual(adder.sub(2, 1), 1)
        self.assertEqual(len(self.cache), 2)
        self.assertEqual(adder.sub(2, 1), 1)

    def test_can_cache_kwargs(self):
        adder = self.DummyClass(self.cache)
        adder.add(x=2, y=1)
        self.assertEqual(adder.add(x=2, y=1), 3)
        self.assertEqual(len(self.cache), 1)


class TestGetServiceModuleName(unittest.TestCase):
    def setUp(self):
        self.service_description = {
            'metadata': {
                'serviceFullName': 'CDP MyService',
                'apiVersion': '2014-01-01',
                'endpointPrefix': 'myservice',
                'signatureVersion': 'v4',
                'protocol': 'query'
            },
            'operations': {},
            'shapes': {},
        }
        self.service_model = ServiceModel(
            self.service_description, 'myservice')

    def test_default(self):
        self.assertEqual(
            get_service_module_name(self.service_model),
            'myservice'
        )

    def test_client_name_with_cloudera(self):
        self.service_description['metadata']['serviceFullName'] = (
            'Cloudera MyService')
        self.assertEqual(
            get_service_module_name(self.service_model),
            'myservice'
        )

    def test_client_name_with_non_alphabet_characters(self):
        self.service_description['metadata']['serviceFullName'] = (
            'Cloudera My-Service')
        self.assertEqual(
            get_service_module_name(self.service_model),
            'myservice'
        )

    def test_client_name_with_no_full_name_or_abbreviation(self):
        del self.service_description['metadata']['serviceFullName']
        self.assertEqual(
            get_service_module_name(self.service_model),
            'myservice'
        )


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli.clidriver import CLIDriver, ServiceOperation
from cdpcli.client import ClientCreator
from cdpcli.exceptions import ExtensionImportError
from cdpcli.model import ServiceModel
from cdpcli.parser import ResponseParserFactory
from mock import Mock, patch
from tests import unittest
import yaml


MODEL_DIR = os.path.dirname(os.path.abspath(__file__))


class TestOperationExtension(unittest.TestCase):

    def setUp(self):
        self.model = yaml.safe_load(open(os.path.join(MODEL_DIR, 'service.yaml')))
        self.service_model = ServiceModel(self.model, 'servicename')
        self.loader = Mock()
        self.loader.load_service_data.return_value = self.model
        self.endpoint = Mock()
        self.endpoint.host = 'http://thunderhead.cloudera.altus.cloudera.com'
        self.endpoint.make_request.return_value = (Mock(status_code=200), {})
        self.endpoint_creator = Mock()
        self.endpoint_creator.create_endpoint.return_value = self.endpoint
        self.context = Mock()
        self.context.get_credentials.return_value = Mock(private_key="A private key")
        self.context.get_scoped_config.return_value = {}
        self.client_creator = ClientCreator(self.loader,
                                            self.context,
                                            self.endpoint_creator,
                                            'user-agent-header',
                                            ResponseParserFactory(),
                                            Mock())
        self.parsed_globals = Mock(connect_timeout=1, read_timeout=1)

    def test_invalid_operation_extension(self):
        operation_caller = Mock()
        service_operation = ServiceOperation(
            Mock(spec=CLIDriver),
            'create-director',
            'directors',
            operation_caller,
            self.service_model.operation_model('createDirector'))
        with self.assertRaises(ExtensionImportError) as context:
            service_operation(self.client_creator, [], self.parsed_globals)
        self.assertTrue(
            'Failed to import CLI extension \'invalid\': '
            'Not Found' in str(context.exception))
        self.assertFalse(operation_caller.invoke.called)

    def test_operation_replace_by_extension(self):
        extension_caller = Mock()
        extension_caller.invoke.return_value = False
        test_obj = Mock()
        test_obj.register = lambda operation_callers, operation_model: \
            operation_callers.insert(0, extension_caller)
        module_obj = {'cdpcli.extensions.test': test_obj}

        with patch('importlib.import_module', new=module_obj.get):
            operation_caller = Mock()
            service_operation = ServiceOperation(
                Mock(spec=CLIDriver),
                'upload-data',
                'directors',
                operation_caller,
                self.service_model.operation_model('uploadData'))
            service_operation(self.client_creator, [], self.parsed_globals)
            self.assertFalse(operation_caller.invoke.called)

        self.assertTrue(extension_caller.invoke.called)
        self.assertEquals(1, extension_caller.invoke.call_count)

    def test_operation_extensions_injected(self):
        extension_call_list = []
        extension_caller_1 = Mock()
        extension_caller_1.invoke.side_effect = lambda *args, **kwargs: \
            extension_call_list.append(1)
        extension_caller_1.invoke.return_value = True
        extension_caller_2 = Mock()
        extension_caller_2.invoke.side_effect = lambda *args, **kwargs: \
            extension_call_list.append(2)
        extension_caller_2.invoke.return_value = True
        test_obj_1 = Mock()
        test_obj_1.register = lambda operation_callers, operation_model: \
            operation_callers.insert(0, extension_caller_1)
        test_obj_2 = Mock()
        test_obj_2.register = lambda operation_callers, operation_model: \
            operation_callers.insert(0, extension_caller_2)
        module_obj = {'cdpcli.extensions.test1': test_obj_1,
                      'cdpcli.extensions.test2': test_obj_2}

        with patch('importlib.import_module', new=module_obj.get):
            operation_caller = Mock()
            service_operation = ServiceOperation(
                Mock(spec=CLIDriver),
                'inject-data',
                'directors',
                operation_caller,
                self.service_model.operation_model('injectData'))
            service_operation(self.client_creator, [], self.parsed_globals)
            self.assertTrue(operation_caller.invoke.called)
            self.assertEquals(1, operation_caller.invoke.call_count)

        self.assertTrue(extension_caller_1.invoke.called)
        self.assertEquals(1, extension_caller_1.invoke.call_count)
        self.assertTrue(extension_caller_2.invoke.called)
        self.assertEquals(1, extension_caller_2.invoke.call_count)
        self.assertEquals([1, 2], extension_call_list)

    def test_operation_extensions_update_parsed_globals(self):
        # noinspection PyShadowingNames
        def _invoke_to_update_parsed_globals(client_creator,
                                             operation_model,
                                             parameters,
                                             parsed_args,
                                             parsed_globals):
            service_name = operation_model.service_model.service_name
            client_creator(service_name)
            parsed_globals.endpoint_url = 'https://test.com/'
            parsed_globals.access_token = 'Bearer ABC'
            return True

        # noinspection PyShadowingNames
        def _mock_operation_caller_invoke(client_creator,
                                          operation_model,
                                          parameters,
                                          parsed_args,
                                          parsed_globals):
            service_name = operation_model.service_model.service_name
            client_creator(service_name)

        extension_caller = Mock()
        extension_caller.invoke = _invoke_to_update_parsed_globals
        test_obj = Mock()
        test_obj.register = lambda operation_callers, operation_model: \
            operation_callers.insert(0, extension_caller)
        module_obj = {'cdpcli.extensions.test': test_obj}

        with patch('importlib.import_module', new=module_obj.get):
            operation_caller = Mock()
            operation_caller.invoke.side_effect = _mock_operation_caller_invoke
            service_operation = ServiceOperation(
                Mock(spec=CLIDriver),
                'upload-data',
                'directors',
                operation_caller,
                self.service_model.operation_model('uploadData'))
            service_operation(self.client_creator, [], self.parsed_globals)
            self.assertTrue(operation_caller.invoke.called)
            self.assertEquals(1, operation_caller.invoke.call_count)
            args, kargs = operation_caller.invoke.call_args
            client_creator, operation_model, parameters, parsed_args, parsed_globals = \
                args
            self.assertEquals('uploadData', operation_model.name)
            self.assertEquals('https://test.com/', parsed_globals.endpoint_url)
            self.assertEquals('Bearer ABC', parsed_globals.access_token)
            args, kargs = self.endpoint_creator.create_endpoint.call_args
            service_model, endpoint_url, scoped_config, region, response_parser_factory, \
                tls_verification, timeout, retry_handler = args
            self.assertEquals('https://test.com/', endpoint_url)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import base64
import datetime
import decimal

from cdpcli.compat import six
from cdpcli.exceptions import ParamValidationError
from cdpcli.utils import parse_to_aware_datetime


def validate_parameters(params, shape):
    """Validates input parameters against a schema.

    This is a convenience function that validates parameters against a schema.
    You can also instantiate and use the ParamValidator class directly if you
    want more control.

    If there are any validation errors then a ParamValidationError
    will be raised.  If there are no validation errors than no exception
    is raised and a value of None is returned.
    """
    validator = ParamValidator()
    report = validator.validate(params, shape)
    if report.has_errors():
        raise ParamValidationError(report=report.generate_report())


def type_check(valid_types):
    def _create_type_check_guard(func):
        def _on_passes_type_check(self, param, shape, errors, name):
            if _type_check(param, errors, name):
                return func(self, param, shape, errors, name)

        def _type_check(param, errors, name):
            if not isinstance(param, valid_types):
                valid_type_names = [six.text_type(t) for t in valid_types]
                errors.report(name, 'invalid type', param=param,
                              valid_types=valid_type_names)
                return False
            return True

        return _on_passes_type_check
    return _create_type_check_guard


def range_check(name, value, shape, error_type, errors):
    failed = False
    min_allowed = float('-inf')
    max_allowed = float('inf')
    if shape.minimum is not None:
        min_allowed = shape.minimum
        if value < min_allowed:
            failed = True
    if shape.maximum is not None:
        max_allowed = shape.maximum
        if value > max_allowed:
            failed = True
    if failed:
        errors.report(name, error_type, param=value,
                      valid_range=[min_allowed, max_allowed])


def length_check(name, value, shape, error_type, errors):
    failed = False
    min_allowed = float('-inf')
    max_allowed = float('inf')
    if shape.min_length is not None:
        min_allowed = shape.min_length
        if value < min_allowed:
            failed = True
    if shape.max_length is not None:
        max_allowed = shape.max_length
        if value > max_allowed:
            failed = True
    if failed:
        errors.report(name, error_type, param=value,
                      valid_range=[min_allowed, max_allowed])


class ValidationErrors(object):
    def __init__(self):
        self._errors = []

    def has_errors(self):
        if self._errors:
            return True
        return False

    def generate_report(self):
        error_messages = []
        for error in self._errors:
            error_messages.append(self._format_error(error))
        return '\n'.join(error_messages)

    def _format_error(self, error):
        error_type, name, additional = error
        name = self._get_name(name)
        if error_type == 'missing required field':
            return 'Missing required parameter in %s: "%s"' % (
                name, additional['required_name'])
        elif error_type == 'unknown field':
            return 'Unknown parameter in %s: "%s", must be one of: %s' % (
                name, additional['unknown_param'], ', '.join(additional['valid_names']))
        elif error_type == 'invalid type':
            return 'Invalid type for parameter %s, value: %s, type: %s, valid types: %s' \
                % (name, additional['param'],
                   str(type(additional['param'])),
                   ', '.join(additional['valid_types']))
        elif error_type == 'invalid enum':
            return ('Invalid value for parameter %s, value: %s, type: %s, valid '
                    'values: %s') \
                % (name, additional['param'],
                   str(type(additional['param'])),
                   ', '.join(additional['valid_values']))
        elif error_type == 'invalid range':
            min_allowed = additional['valid_range'][0]
            max_allowed = additional['valid_range'][1]
            return ('Invalid range for parameter %s, value: %s, valid range: '
                    '%s-%s' % (name, additional['param'],
                               min_allowed, max_allowed))
        elif error_type == 'invalid length':
            min_allowed = additional['valid_range'][0]
            max_allowed = additional['valid_range'][1]
            return ('Invalid length for parameter %s, value: %s, valid range: '
                    '%s-%s' % (name, additional['param'],
                               min_allowed, max_allowed))
        elif error_type == 'invalid base64':
            return 'Invalid base64 value for parameter %s, error: %s' \
                % (name, additional['error'])

    def _get_name(self, name):
        if not name:
            return 'input'
        elif name.startswith('.'):
            return name[1:]
        else:
            return name

    def report(self, name, reason, **kwargs):
        self._errors.append((reason, name, kwargs))


class ParamValidator(object):

    def validate(self, params, shape):
        errors = ValidationErrors()
        self._validate(params, shape, errors, name='')
        return errors

    def _validate(self, params, shape, errors, name):
        getattr(self, '_validate_%s' % shape.type_name)(params, shape, errors, name)

    @type_check(valid_types=(dict,))
    def _validate_object(self, params, shape, errors, name):
        # Validate required fields.
        for required_member in shape.required_members:
            if required_member not in params:
                errors.report(name, 'missing required field',
                              required_name=required_member, user_params=params)
        members = shape.members
        known_params = []
        # Validate known params.
        for param in params:
            if param not in members:
                errors.report(name, 'unknown field', unknown_param=param,
                              valid_names=list(members))
            else:
                known_params.append(param)
        # Validate structure members.
        for param in known_params:
            self._validate(params[param], shape.members[param],
                           errors, '%s.%s' % (name, param))

    @type_check(valid_types=(dict,))
    def _validate_map(self, params, shape, errors, name):
        for key, value in params.items():
            self._validate(key, shape.key, errors, "%s (key: %s)"
                           % (name, key))
            self._validate(value, shape.value, errors, '%s.%s' % (name, key))

    @type_check(valid_types=(bool,))
    def _validate_boolean(self, param, shape, errors, name):
        pass

    @type_check(valid_types=six.integer_types)
    def _validate_integer(self, param, shape, errors, name):
        range_check(name, param, shape, 'invalid range', errors)

    @type_check(valid_types=(float, decimal.Decimal) + six.integer_types)
    def _validate_float(self, param, shape, errors, name):
        range_check(name, param, shape, 'invalid range', errors)

    @type_check(valid_types=(float, decimal.Decimal) + six.integer_types)
    def _validate_double(self, param, shape, errors, name):
        range_check(name, param, shape, 'invalid range', errors)

    @type_check(valid_types=six.string_types)
    def _validate_string(self, param, shape, errors, name):
        length_check(name, len(param), shape, 'invalid length', errors)

    @type_check(valid_types=(list, tuple))
    def _validate_array(self, param, shape, errors, name):
        member_shape = shape.member
        length_check(name, len(param), shape, 'invalid length', errors)
        for i, item in enumerate(param):
            self._validate(item, member_shape, errors, '%s[%s]' % (name, i))

    def _validate_datetime(self, param, shape, errors, name):
        # We don't use @type_check because datetimes are a bit more flexible.
        # You can either provide a datetime object, or a string that parses
        # to a datetime.
        is_valid_type = self._type_check_datetime(param)
        if not is_valid_type:
            valid_type_names = [six.text_type(datetime), 'timestamp-string']
            errors.report(name, 'invalid type', param=param,
                          valid_types=valid_type_names)

    @type_check(valid_types=(bytes, bytearray, six.string_types))
    def _validate_blob(self, param, shape, errors, name):
        if isinstance(param, six.string_types):
            try:
                bytes = base64.b64decode(param)
                length_check(name, len(bytes), shape, 'invalid length', errors)
            except Exception as err:
                errors.report(name, 'invalid base64', error=err)
        else:
            length_check(name, len(param), shape, 'invalid length', errors)

    def _type_check_datetime(self, value):
        try:
            parse_to_aware_datetime(value)
            return True
        except (TypeError, ValueError, AttributeError):
            # Yes, dateutil can sometimes raise an AttributeError when parsing
            # timestamps.
            return False


class ParamValidationDecorator(object):
    def __init__(self, param_validator, serializer):
        self._param_validator = param_validator
        self._serializer = serializer

    def serialize_to_request(self, parameters, operation_model):
        input_shape = operation_model.input_shape
        if input_shape is not None:
            report = self._param_validator.validate(parameters,
                                                    operation_model.input_shape)
            if report.has_errors():
                raise ParamValidationError(report=report.generate_report())
        return self._serializer.serialize_to_request(parameters, operation_model)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from types import MethodType

from cdpcli.exceptions import PaginationError
from cdpcli.model import ObjectShape
from cdpcli.model import OperationModel
from cdpcli.paginate import Paginator
from mock import call
from mock import Mock
from tests import unittest


class TestPagination(unittest.TestCase):

    def setUp(self):
        self.method = Mock(spec=MethodType)
        self.operation_model = Mock(spec=OperationModel)
        self.operation_model.paging_input_token = 'NextToken'
        self.operation_model.paging_output_token = 'NextToken'
        self.operation_model.paging_result = 'Foo'
        self.operation_model.paging_page_size = 'pageSize'
        self.operation_model.input_shape = Mock(spec=ObjectShape)
        self.operation_model.input_shape.members = {'pageSize': Mock()}
        self.operation_model.input_shape.members['pageSize'].maximum = 100
        self.operation_model.paging_default_max_items = 1000
        self.paginator = Paginator(self.method, self.operation_model)

    def test_no_next_token(self):
        response = {'not_the_next_token': 'foobar'}
        self.method.return_value = response
        actual = list(self.paginator.paginate())
        self.assertEqual(actual, [{'not_the_next_token': 'foobar'}])

    def test_next_token_in_response(self):
        responses = [{'NextToken': 'token1'},
                     {'NextToken': 'token2'},
                     {'not_next_token': 'foo'}]
        self.method.side_effect = responses
        actual = list(self.paginator.paginate())
        self.assertEqual(actual, responses)
        self.assertEqual(self.method.call_args_list, [call(pageSize=100),
                                                      call(NextToken='token1',
                                                           pageSize=100),
                                                      call(NextToken='token2',
                                                           pageSize=100)])

    def test_build_full_result(self):
        responses = [{'Foo': ['User1'], 'NextToken': 'm1'},
                     {'Foo': ['User2'], 'NextToken': 'm2'},
                     {'Foo': ['User3']}]
        self.method.side_effect = responses
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})

    def test_build_full_result_no_result(self):
        responses = [{'Foo': ['User1'], 'NextToken': 'm1'},
                     {'NextToken': 'm2'},
                     {'Foo': ['User2']}]
        self.method.side_effect = responses
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2']})

    def test_build_full_result_sum(self):
        responses = [{'Foo': 2.0, 'NextToken': 'm1'},
                     {'NextToken': 'm2'},
                     {'Foo': 4.0}]
        self.method.side_effect = responses
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': 6.0})

    def test_build_full_result_strings(self):
        responses = [{'Foo': 'foo', 'NextToken': 'm1'},
                     {'NextToken': 'm2'},
                     {'Foo': 'bar'}]
        self.method.side_effect = responses
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': 'foobar'})

    def test_exception_raised_if_same_next_token(self):
        responses = [{'NextToken': 'token1'},
                     {'NextToken': 'token2'},
                     {'NextToken': 'token2'}]
        self.method.side_effect = responses
        with self.assertRaises(PaginationError):
            list(self.paginator.paginate())

    def test_build_full_result_max_items(self):
        responses = [{'Foo': ['User1'], 'NextToken': 'm1'},
                     {'Foo': ['User2'], 'NextToken': 'm2'},
                     {'Foo': ['User3']}]
        self.operation_model.paging_default_max_items = 1
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1'], 'nextToken': 'm1'})
        self.method.assert_called_with(pageSize=1)

        self.operation_model.paging_default_max_items = 2
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2'],
                                  'nextToken': 'm2'})
        self.method.assert_any_call(pageSize=2)
        self.method.assert_called_with(pageSize=1, NextToken='m1')

        self.operation_model.paging_default_max_items = 3
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=3)
        self.method.assert_any_call(pageSize=2, NextToken='m1')
        self.method.assert_called_with(pageSize=1, NextToken='m2')

        self.operation_model.paging_default_max_items = 4
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=4)
        self.method.assert_any_call(pageSize=3, NextToken='m1')
        self.method.assert_called_with(pageSize=2, NextToken='m2')

        self.operation_model.paging_default_max_items = 101
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=100)
        self.method.assert_any_call(pageSize=100, NextToken='m1')
        self.method.assert_called_with(pageSize=99, NextToken='m2')

        self.operation_model.paging_default_max_items = 1000
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=100)
        self.method.assert_any_call(pageSize=100, NextToken='m1')
        self.method.assert_called_with(pageSize=100, NextToken='m2')

        # all items at once
        responses = [{'Foo': ['User1', 'User2', 'User3']}]
        self.operation_model.paging_default_max_items = 3
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate().build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_called_with(pageSize=3)

    def test_build_full_result_paging_params(self):
        responses = [{'Foo': ['User1'], 'NextToken': 'm1'},
                     {'Foo': ['User2'], 'NextToken': 'm2'},
                     {'Foo': ['User3']}]

        self.operation_model.paging_default_max_items = 500

        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate(
            PaginationConfig={'MaxItems': 1}).build_full_result()
        self.assertEqual(result, {'Foo': ['User1'], 'nextToken': 'm1'})
        self.method.assert_called_with(pageSize=1)

        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate(
            PaginationConfig={'MaxItems': 2}).build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2'],
                                  'nextToken': 'm2'})
        self.method.assert_any_call(pageSize=2)
        self.method.assert_called_with(pageSize=1, NextToken='m1')

        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate(
            PaginationConfig={'MaxItems': 3}).build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=3)
        self.method.assert_any_call(pageSize=2, NextToken='m1')
        self.method.assert_called_with(pageSize=1, NextToken='m2')

        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate(
            PaginationConfig={'MaxItems': 4}).build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=4)
        self.method.assert_any_call(pageSize=3, NextToken='m1')
        self.method.assert_called_with(pageSize=2, NextToken='m2')

        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate(
            PaginationConfig={'MaxItems': 101,
                              'PageSize': 50}).build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=50)
        self.method.assert_any_call(pageSize=50, NextToken='m1')
        self.method.assert_called_with(pageSize=50, NextToken='m2')

        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate(
            PaginationConfig={'MaxItems': 1000,
                              'PageSize': 500}).build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=100)
        self.method.assert_any_call(pageSize=100, NextToken='m1')
        self.method.assert_called_with(pageSize=100, NextToken='m2')

        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        result = self.paginator.paginate(
            PaginationConfig={'MaxItems': 101,
                              'PageSize': 50,
                              'StartingToken': 'm0'}).build_full_result()
        self.assertEqual(result, {'Foo': ['User1', 'User2', 'User3']})
        self.method.assert_any_call(pageSize=50, NextToken='m0')
        self.method.assert_any_call(pageSize=50, NextToken='m1')
        self.method.assert_called_with(pageSize=50, NextToken='m2')

    # make sure we raise an error if service sends too many items
    def test_build_full_result_max_items_overshoot(self):
        # one page
        self.operation_model.paging_default_max_items = 2
        responses = [{'Foo': ['User1', 'User2', 'User3']}]
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        with self.assertRaises(PaginationError):
            self.paginator.paginate().build_full_result()
        self.method.assert_called_with(pageSize=2)
        # head
        self.operation_model.paging_default_max_items = 1
        responses = [{'Foo': ['User1', 'User2'], 'NextToken': 'm1'},
                     {'Foo': ['User3', 'User4'], 'NextToken': 'm2'},
                     {'Foo': ['User5', 'User6']}]
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        with self.assertRaises(PaginationError):
            self.paginator.paginate().build_full_result()
        self.method.assert_called_with(pageSize=1)
        # tail
        responses = [{'Foo': ['User1', 'User2'], 'NextToken': 'm1'},
                     {'Foo': ['User3', 'User4'], 'NextToken': 'm2'},
                     {'Foo': ['User5', 'User6']}]
        self.operation_model.paging_default_max_items = 5
        self.paginator = Paginator(self.method, self.operation_model)
        self.method.side_effect = responses
        self.method.reset_mock()
        with self.assertRaises(PaginationError):
            self.paginator.paginate().build_full_result()
        self.method.assert_called_with(pageSize=1, NextToken='m2')

    #  violate Paginator preconditions
    def test_paginator_preconditions(self):
        def feed_bad_values(values, fn):
            for value in values:
                with self.assertRaises(AssertionError):
                    fn(value)
        self.operation_model.paging_default_max_items = 1
        feed_bad_values(
            [None, 1, 'abc', {'abc': 1}, (1, 2), ['a', 'b']],
            lambda value: Paginator(value, self.operation_model))
        feed_bad_values(
            [None, 1, 'abc', {'abc': 1}, (1, 2), ['a', 'b'], feed_bad_values, eval],
            lambda value: Paginator(self.method, value))



# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli.compat import json
from cdpcli.compat import OrderedDict
from cdpcli.endpoint import DEFAULT_TIMEOUT, Endpoint
from cdpcli.endpoint import EndpointCreator
from cdpcli.endpoint import EndpointResolver
from cdpcli.exceptions import EndpointConnectionError
from cdpcli.retryhandler import create_retry_handler
from cdpcli.translate import build_retry_config
from mock import Mock, patch
from requests import ConnectionError
from tests import unittest

CLIENT_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'client')


def request_dict():
    return {
        'headers': {},
        'body': '',
        'url_path': '/',
        'query_string': '',
        'method': 'POST',
        'url': 'https://example.com'
    }


class TestEndpointBase(unittest.TestCase):

    def setUp(self):
        self.op = Mock()
        self.op.metadata = {'protocol': 'json'}
        self.request_signer = Mock()
        self.factory_patch = patch(
            'cdpcli.parser.ResponseParserFactory')
        self.factory = self.factory_patch.start()
        self.retry_handler = Mock(wraps=self._create_retry_handler())
        self.endpoint = Endpoint('https://iamapi.us-west-1.cdp.cloudera.com/',
                                 retry_handler=self.retry_handler)
        self.http_session = Mock()
        self.http_session.send.return_value = Mock(
            status_code=200, headers={}, content=b'{"Foo": "bar"}',
        )
        self.endpoint.http_session = self.http_session

    def tearDown(self):
        self.factory_patch.stop()

    def _create_retry_handler(self):
        original_config = open(os.path.join(CLIENT_DIR, '_retry.json')).read()
        original_config = json.loads(original_config,
                                     object_pairs_hook=OrderedDict)
        config = build_retry_config(
            original_config['retry'],
            original_config.get('definitions', {}))
        return create_retry_handler(config)


class TestEndpointFeatures(TestEndpointBase):

    def test_timeout_can_be_specified(self):
        timeout_override = 120
        self.endpoint.timeout = timeout_override
        self.endpoint.make_request(self.op, request_dict(), self.request_signer)
        kwargs = self.http_session.send.call_args[1]
        self.assertEqual(kwargs['timeout'], timeout_override)

    def test_make_request(self):
        self.endpoint.make_request(self.op, request_dict(), self.request_signer)
        prepared_request = self.http_session.send.call_args[0][0]
        self.http_session.send.assert_called_with(
            prepared_request, verify=True, stream=False,
            proxies={}, timeout=DEFAULT_TIMEOUT, allow_redirects=True)

    def test_make_request_with_no_redirects(self):
        self.endpoint.make_request(self.op, request_dict(), self.request_signer,
                                   allow_redirects=False)
        prepared_request = self.http_session.send.call_args[0][0]
        self.http_session.send.assert_called_with(
            prepared_request, verify=True, stream=False,
            proxies={}, timeout=DEFAULT_TIMEOUT, allow_redirects=False)

    def test_make_request_with_proxies(self):
        proxies = {'http': 'http://localhost:8888'}
        self.endpoint.proxies = proxies
        self.endpoint.make_request(self.op, request_dict(), self.request_signer)
        prepared_request = self.http_session.send.call_args[0][0]
        self.http_session.send.assert_called_with(
            prepared_request, verify=True, stream=False,
            proxies=proxies, timeout=DEFAULT_TIMEOUT, allow_redirects=True)

    def test_make_request_with_no_auth(self):
        self.endpoint.auth = None
        self.endpoint.make_request(self.op, request_dict(), self.request_signer)

        # http_session should be used to send the request.
        self.assertTrue(self.http_session.send.called)
        prepared_request = self.http_session.send.call_args[0][0]
        self.assertNotIn('Authorization', prepared_request.headers)

    def test_make_request_no_signature_version(self):
        self.endpoint.make_request(self.op, request_dict(), self.request_signer)

        # http_session should be used to send the request.
        self.assertTrue(self.http_session.send.called)
        prepared_request = self.http_session.send.call_args[0][0]
        self.assertNotIn('Authorization', prepared_request.headers)

    def test_make_request_injects_better_dns_error_msg(self):
        fake_request = Mock(url='https://iamapi.us-west-1.altus.cloudera.com')
        self.http_session.send.side_effect = ConnectionError(
            "Fake gaierror(8, node or host not known)", request=fake_request)
        with self.assertRaisesRegexp(EndpointConnectionError,
                                     'Could not connect'):
            self.endpoint.make_request(self.op, request_dict(),
                                       self.request_signer)

    def test_make_request_injects_better_bad_status_line_error_msg(self):
        fake_request = Mock(url='https://iamapi.us-west-1.altus.cloudera.com')
        self.http_session.send.side_effect = ConnectionError(
            """'Connection aborted.', BadStatusLine("''",)""",
            request=fake_request)
        with self.assertRaisesRegexp(ConnectionError,
                                     'Connection aborted.'):
            self.endpoint.make_request(self.op, request_dict(),
                                       self.request_signer)


class TestRetryInterface(TestEndpointBase):
    def setUp(self):
        super(TestRetryInterface, self).setUp()
        self.retried_on_exception = None
        self.total_calls = 0

    def test_retry_is_attempted(self):
        op = Mock()
        op.name = 'DescribeCluster'
        op.metadata = {'protocol': 'query'}
        self.endpoint.make_request(op, request_dict(), self.request_signer)
        call_args = self.endpoint._retry_handler.call_args
        self.assertEqual(call_args[1]['attempts'], 1)

    def test_retry_events_can_alter_behavior(self):
        op = Mock()
        op.name = 'DescribeCluster'
        op.metadata = {'protocol': 'json'}
        self.retry_handler.side_effect = [
            0,       # Check if retry needed. Retry needed.
            None     # Check if retry needed. Retry not needed.
        ]
        self.endpoint.make_request(op, request_dict(), self.request_signer)
        self.assertEqual(self.retry_handler.call_count, 2)

    def test_retry_on_socket_errors(self):
        op = Mock()
        op.name = 'DescribeClusters'
        self.retry_handler.side_effect = [
            0,       # Check if retry needed. Retry needed.
            None     # Check if retry needed. Retry not needed.
        ]
        self.http_session.send.side_effect = ConnectionError()
        with self.assertRaises(ConnectionError):
            self.endpoint.make_request(op, request_dict(), self.request_signer)
        self.assertEqual(self.retry_handler.call_count, 2)

    def test_retry_on_429(self):
        http_429_response = Mock()
        http_429_response.status_code = 429
        http_429_response.headers = []
        http_429_response.content = b'{}'
        op = Mock()
        op.name = 'DescribeClusters'
        self.http_session.send.return_value = http_429_response
        self.endpoint.make_request(op, request_dict(), self.request_signer)
        self.assertEqual(self.retry_handler.call_count, 3)

    def test_retry_on_503(self):
        http_503_response = Mock()
        http_503_response.status_code = 503
        http_503_response.headers = []
        http_503_response.content = b'{}'
        op = Mock()
        op.name = 'DescribeClusters'
        self.http_session.send.return_value = http_503_response
        self.endpoint.make_request(op, request_dict(), self.request_signer)
        self.assertEqual(self.retry_handler.call_count, 3)

    def test_retry_on_509(self):
        http_509_response = Mock()
        http_509_response.status_code = 509
        http_509_response.headers = []
        http_509_response.content = b'{}'
        op = Mock()
        op.name = 'DescribeClusters'
        self.http_session.send.return_value = http_509_response
        self.endpoint.make_request(op, request_dict(), self.request_signer)
        self.assertEqual(self.retry_handler.call_count, 3)

    def test_no_retry_on_success(self):
        http_200_response = Mock()
        http_200_response.status_code = 200
        http_200_response.headers = []
        http_200_response.content = b'{}'
        op = Mock()
        op.name = 'DescribeClusters'
        self.http_session.send.return_value = http_200_response
        self.endpoint.make_request(op, request_dict(), self.request_signer)
        self.assertEqual(self.retry_handler.call_count, 1)


class TestEndpointCreator(unittest.TestCase):
    def setUp(self):
        self.service_model = Mock(
            endpoint_name='iam',
            signature_version='v1',
            signing_name='iam')
        self.cdp_service_model = Mock(
            endpoint_name='datahub',
            endpoint_prefix='api',
            signature_version='v1',
            signing_name='datahub',
            products=['CDP'])
        self.environ = {}
        self.environ_patch = patch('os.environ', self.environ)
        self.environ_patch.start()
        self.creator = EndpointCreator(EndpointResolver())
        self.factory_patch = patch(
            'cdpcli.parser.ResponseParserFactory')
        self.factory = self.factory_patch.start()
        self.retry_handler = Mock()

    def tearDown(self):
        self.environ_patch.stop()

    def test_altus_endpoint(self):
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=None,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, 'https://iamapi.us-west-1.altus.cloudera.com:443')

    def test_cdp_endpoint(self):
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, 'https://api.us-west-1.cdp.cloudera.com:443')

    def test_cdp_endpoint_with_prefix(self):
        self.cdp_service_model.endpoint_prefix = 'prefix'
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host,
                         'https://prefix.us-west-1.cdp.cloudera.com:443')

    def test_create_endpoint_with_configured_url(self):
        endpoint_url = 'https://endpoint.url'
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=endpoint_url,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_customized_timeout(self):
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url='https://example.com',
            scoped_config={},
            timeout=123,
            region='default',
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.timeout, 123)

    def test_create_endpoint_with_config_url_altus(self):
        endpoint_url = 'https://endpoint-config.url'
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=None,
            scoped_config={"endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_config_url_cdp(self):
        endpoint_url = 'https://endpoint-config.url'
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={"cdp_endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_configured_url_substitution(self):
        endpoint_url = 'https://%s.endpoint.url'
        expected = endpoint_url % self.service_model.endpoint_name
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=endpoint_url,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, expected)

    def test_create_endpoint_with_configured_cdp_url_substitution(self):
        endpoint_url = 'https://%s.endpoint.url'
        self.cdp_service_model.endpoint_prefix = 'prefix'
        expected = endpoint_url % self.cdp_service_model.endpoint_prefix
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=endpoint_url,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, expected)

    def test_create_endpoint_with_config_url_substitution(self):
        endpoint_url = 'https://%s.endpoint-config.url'
        expected = endpoint_url % self.service_model.endpoint_name
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=None,
            scoped_config={"endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, expected)

    def test_create_endpoint_with_config_cdp_url_substitution(self):
        endpoint_url = 'https://%s.endpoint-config.url'
        self.cdp_service_model.endpoint_prefix = 'prefix'
        expected = endpoint_url % self.cdp_service_model.endpoint_prefix
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={"cdp_endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, expected)

    def test_create_endpoint_with_configured_url_bad_substitution(self):
        endpoint_url = 'https://%s.%s.endpoint.url'
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=endpoint_url,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_configured_cdp_url_bad_substitution(self):
        endpoint_url = 'https://%s.%s.endpoint.url'
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=endpoint_url,
            scoped_config={},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_config_url_bad_substitution(self):
        endpoint_url = 'https://%s.%s.endpoint-config.url'
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=None,
            scoped_config={"endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_config_cdp_url_bad_substitution(self):
        endpoint_url = 'https://%s.%s.endpoint-config.url'
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={"cdp_endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_for_internal_apis(self):
        endpoint_url = 'https://foo-api.internal.url'
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={"cdp_endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_wildcard_endpoint_for_internal_apis(self):
        endpoint_url = 'https://%s-api.internal.url'
        self.cdp_service_model.endpoint_prefix = 'prefix'
        expected = endpoint_url % self.cdp_service_model.endpoint_prefix
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={"cdp_endpoint_url": endpoint_url,
                           "config1": "value1"},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, expected)

    def test_create_endpoint_with_explicit_region_us_west_1(self):
        endpoint_url = 'https://iamapi.us-west-1.altus.cloudera.com:443'
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=None,
            scoped_config={},
            region='us-west-1',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_explicit_region_eu_1(self):
        endpoint_url = 'https://api.eu-1.cdp.cloudera.com:443'
        endpoint = self.creator.create_endpoint(
            self.service_model,
            explicit_endpoint_url=None,
            scoped_config={},
            region='eu-1',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_region_from_config_us_west_1(self):
        endpoint_url = 'https://api.us-west-1.cdp.cloudera.com:443'
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={'cdp_region': 'us-west-1'},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)

    def test_create_endpoint_with_region_from_config_eu_1(self):
        endpoint_url = 'https://api.eu-1.cdp.cloudera.com:443'
        endpoint = self.creator.create_endpoint(
            self.cdp_service_model,
            explicit_endpoint_url=None,
            scoped_config={'cdp_region': 'eu-1'},
            region='default',
            timeout=123,
            response_parser_factory=self.factory_patch,
            tls_verification=False,
            retry_handler=self.retry_handler)
        self.assertEqual(endpoint.host, endpoint_url)


class TestEndpointResolver(unittest.TestCase):
    def test_resolve_login_endpoint(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={},
            region=None,
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')

    def test_resolve_login_endpoint_with_explicit_url(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url="http://explict.url/path?accountId=foo",
            config={},
            region=None,
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint, 'http://explict.url/path?accountId=foo')

    def test_resolve_login_endpoint_with_login_url_from_config(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={'login_url': 'http://config.url/path?accountId=foo'},
            region='eu-1',  # region is be ignored when using 'login_url' from config
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint, 'http://config.url/path?accountId=foo')

    def test_resolve_login_endpoint_with_explicit_region_default(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={},
            region='default',
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')

    def test_resolve_login_endpoint_with_explicit_region_us_west_1(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={},
            region='us-west-1',
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')

    def test_resolve_login_endpoint_with_explicit_region_eu_1(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={},
            region='eu-1',
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint,
                         'https://console.eu-1.cdp.cloudera.com:443/consoleauth/login')

    def test_resolve_login_endpoint_with_region_from_config_default(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={'cdp_region': 'default'},
            region=None,
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')

    def test_resolve_login_endpoint_with_region_from_config_us_west_1(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={'cdp_region': 'us-west-1'},
            region=None,
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')

    def test_resolve_login_endpoint_with_region_from_config_eu_1(self):
        endpoint_resolver = EndpointResolver()
        endpoint = endpoint_resolver.resolve(
            explicit_endpoint_url=None,
            config={'cdp_region': 'eu-1'},
            region=None,
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)
        self.assertEqual(endpoint,
                         'https://console.eu-1.cdp.cloudera.com:443/consoleauth/login')



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli.compat import six
from cdpcli.extensions.generatecliskeleton import add_generate_skeleton
from cdpcli.extensions.generatecliskeleton import GenerateCliSkeletonArgument
from cdpcli.model import DenormalizedStructureBuilder
import mock
from tests import unittest

from . import FakeArgs


class TestGenerateCliSkeleton(unittest.TestCase):
    def setUp(self):
        # Create a mock service operation object
        self.service_operation = mock.Mock()

        # Make an arbitrary input model shape.
        self.input_shape = {
            'A': {
                'type': 'object',
                'properties': {
                    'B': {'type': 'string'},
                }
            }
        }
        shape = DenormalizedStructureBuilder().with_members(
            self.input_shape).build_model()
        self.operation_model = mock.Mock(input_shape=shape)
        self.argument = GenerateCliSkeletonArgument(self.operation_model)

        # This is what the json should should look like after being
        # generated to standard output.
        self.ref_json_output = \
            '{\n    "A": {\n        "B": ""\n    }\n}\n'

    def test_add_argument_action(self):
        argument_table = {}
        add_generate_skeleton(self.operation_model, argument_table)
        arg_handler = argument_table['generate-cli-skeleton']
        self.assertIsNotNone(arg_handler)
        self.assertTrue(isinstance(arg_handler, GenerateCliSkeletonArgument))

    def test_generate_json_skeleton(self):
        parsed_args = mock.Mock()
        parsed_args.generate_cli_skeleton = True
        with mock.patch('sys.stdout', six.StringIO()) as mock_stdout:
            rc = self.argument._generate_json_skeleton(
                parsed_args=parsed_args
            )
            # Ensure the contents printed to standard output are correct.
            self.assertEqual(self.ref_json_output, mock_stdout.getvalue())
            # Ensure it is the correct return code of zero.
            self.assertEqual(rc, 0)

    def test_no_generate_json_skeleton(self):
        parsed_args = FakeArgs(generate_cli_skeleton=False)
        with mock.patch('sys.stdout', six.StringIO()) as mock_stdout:
            rc = self.argument._generate_json_skeleton(
                parsed_args=parsed_args
            )
            # Ensure nothing is printed to standard output
            self.assertEqual('', mock_stdout.getvalue())
            # Ensure True is returned because it was never called.
            self.assertEqual(rc, True)

    def test_generate_json_skeleton_no_input_shape(self):
        parsed_args = FakeArgs(generate_cli_skeleton=True)
        # Set the input shape to ``None``.
        self.argument = GenerateCliSkeletonArgument(mock.Mock(input_shape=None))
        with mock.patch('sys.stdout', six.StringIO()) as mock_stdout:
            rc = self.argument._generate_json_skeleton(
                parsed_args=parsed_args
            )
            # Ensure the contents printed to standard output are correct,
            # which should be an empty dictionary.
            self.assertEqual('{}\n', mock_stdout.getvalue())
            # Ensure False is returned because it was called.
            self.assertEqual(rc, False)
